# Heat Integration: Stream Definition and Pinch Analysis by TI Method
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2017
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass, field
from typing import Literal

StreamKind = Literal['hot', 'cold']


@dataclass
class Stream:
    """
    Container for a stream in a heat integration problem.

    Examples:
        H1 = Stream(name='H1', kind='hot', T_source=180, T_target=60, H=360)
        C1 = Stream(name='C1', kind='cold', T_source=20, T_target=135, H=230)

    Args:
        name: A descriptive name for the stream.
        kind: Kind (type) of stream. Options are ``hot`` and ``cold``.
        T_source: Source temperature (°C).
        T_target: Target temperature (°C).
        H: Enthalpy requirement (kW). Either one of ``H`` or ``C`` must be
            specified.
        C: Extensive heat capacity (kW/°C). Either one of ``H`` or ``C``
            must be specified.
        h: Individual film heat transfer coefficient (J/(s m^2 °C)). Defaults
            to 1.0 J/(s m^2 °C).

    Returns:
        A Stream object containing the specified properties of the stream.
    """

    name: str
    kind: StreamKind
    T_source: float
    T_target: float
    # Enthalpy requirement
    H: float | None = field(
        default=None,
        repr=lambda v: f'{float(v):.1f}' if v is not None else 'None'
    )
    # Extensive heat capacity
    C: float | None = field(
        default=None,
        repr=lambda v: f'{float(v):.2f}' if v is not None else 'None'
    )
    # Film heat transfer coefficient
    h: float | None = field(
        default=None,
        repr=lambda v: f'{float(v):.2f}' if v is not None else 'None'
    )

    def __post_init__(self):
        if self.C is None and self.H is None:
            raise ValueError(
                'Either extensive heat capacity (C) or enthalpy '
                'requirement (H) must be specified for a stream'
            )
        elif self.C is None:
            self.C = float(self.H / np.abs(self.T_source - self.T_target))
        elif self.H is None:
            self.H = float(self.C * np.abs(self.T_source - self.T_target))
        else:
            self.C = float(self.H / np.abs(self.T_source - self.T_target))
            print(
                "Both H and C are specified; overriding C with calculated "
                f"value (= {self.C})"
            )
            pass

        if self.h is None:
            self.h = 1


def pinch_analysis(streams, DT_min=10, disp=True, GCC=False,
                   fsize=6, font=12, facecolor='white', decimals=2):
    """
    Perform pinch analysis by temperature interval (TI) method.

    Examples:
        perform_pinch_analysis([H1, H2, C1, C2], DT_min=10,
                               disp=False, GCC=False)

        TI, fig, ax = perform_pinch_analysis([H1, H2, C1, C2], DT_min=20,
                                             disp=True, GCC=True)
        display(TI)
        fig

    Args:
        streams: A list of previously created Stream objects.
        DT_min: Minimum approach temperature. Defaults to 10 °C.
        disp: True/False value that determines whether stream data will be
            displayed before performing pinch analysis. Defaults to True.
        GCC: True/False value that determines whether GCC will be drawn.
            Defaults to False.
        fsize: Figure size. Defaults to 6.
        font: Base font. Defaults to 12.
        facecolor: Background color of plot. Defaults to ``white``.
        decimals: Number of decimals used for rounding the columns containing
            stream heat capacities. Other columns are not rounded. Defaults
            to 2.

    Returns:
        A pandas DataFrame containing the results of the pinch analysis
            and if requested, a figure containing the GCC.
    """

    if disp is True:
        df = pd.DataFrame(data={
            'Name': [stream.name for stream in streams],
            'Type': [stream.kind for stream in streams],
            'T_source (°C)': [stream.T_source for stream in streams],
            'T_target (°C)': [stream.T_target for stream in streams],
            'H (kW)': [stream.H for stream in streams],
            'C (kW/°C)': [stream.C for stream in streams]
        })
        print('Stream data:')
        print(df)
        print('\nDesign specs:')
        print(f'\tΔTmin = {DT_min} °C\n')

    hot_streams = [stream for stream in streams if stream.kind == 'hot']
    cold_streams = [stream for stream in streams if stream.kind == 'cold']

    # Use list comprehension to process temperature lists
    T_hot = [
        *(stream.T_source for stream in hot_streams),
        *(stream.T_target for stream in hot_streams)
    ]
    T_hot_adj = [T - DT_min for T in T_hot]
    T_cold = [
        *(stream.T_source for stream in cold_streams),
        *(stream.T_target for stream in cold_streams)
    ]
    T_list = [*T_hot_adj, *T_cold]  # combine
    T_list = list(set(T_list))  # merge duplicates
    T_list.sort(reverse=True)  # sort in descending order

    T_upper = np.array(T_list[:-1])
    T_lower = np.array(T_list[1:])

    intervals = (
        T_upper.astype(str) + ' °C' + ' to ' + T_lower.astype(str) + ' °C'
    )

    # Create pandas dataframe
    TI = pd.DataFrame({
        'T_upper': T_upper,
        'T_lower': T_lower
        }, index=intervals)

    TI.columns = pd.MultiIndex.from_arrays(
        (['T_upper', 'T_lower'], ['°C', '°C'])
    )

    T_upper = np.array(T_list[:-1])
    T_lower = np.array(T_list[1:])
    D_T = T_upper - T_lower
    D_C = np.zeros_like(T_upper, dtype=float)

    # Net heat capacity and residual enthalpy for each temperature interval
    for stream in streams:
        if stream in hot_streams:
            adj = DT_min
            C = stream.C  # hot streams have positive C
        elif stream in cold_streams:
            adj = 0
            C = -stream.C  # cold streams have negative C

        stream_T_upper = max(stream.T_source, stream.T_target)
        stream_T_lower = min(stream.T_source, stream.T_target)
        stream_T_range = (
            (stream_T_upper - adj >= T_upper) &
            (stream_T_lower - adj <= T_lower)
        )

        Ccol = np.zeros_like(T_upper, dtype=float)
        Ccol[stream_T_range] = C
        D_C += Ccol  # add stream heat capacity to net heat capacity column

        Cstr = np.round(Ccol, decimals=decimals).astype(str)
        TI[stream.name, 'kW/°C'] = (
            ['' if float(s) == 0.0 else s for s in Cstr]
        )  # add column for stream

    D_H = D_C * D_T  # net enthalpy

    # Add new columns to dataframe
    TI['ΔT', '°C'] = D_T
    TI['ΔC', 'kW/°C'] = D_C
    TI['ΔH', 'kW'] = D_H

    # Calculate residual and adjusted enthalpies
    TI['Residual ΔH', 'kW'] = TI['ΔH'].cumsum()  # residual enthalpy

    TI.loc['-'] = ['', ''] + [''] * len(streams) + ['', '', '', 0]
    idx = TI.index.tolist()
    idx.insert(0, idx.pop(idx.index('-')))
    TI = TI.reindex(idx)  # add top row

    TI['Adjusted ΔH', 'kW'] = (
        TI['Residual ΔH', 'kW'] - min(TI['Residual ΔH', 'kW'])
    )  # adjusted enthalpy

    # Grand composite curve
    if GCC is True:
        fig, ax = plt.subplots(
            figsize=(fsize, fsize), facecolor=facecolor
        )

        GCC_y = TI['T_lower', '°C'].to_numpy()
        GCC_y[0] = TI['T_upper', '°C'].to_numpy()[1]
        GCC_x = TI['Adjusted ΔH', 'kW'].to_numpy()

        ax.plot(GCC_x, GCC_y, 'k-')
        ax.set_xlabel('H (kW)', fontsize=font)
        ax.set_ylabel('T (°C)', fontsize=font)

    if GCC is True:
        return TI, fig, ax
    else:
        return TI


determine_pinch = find_pinch = perform_pinch_analysis = pinch_analysis
implement_TI = implement_TI_method = pinch_analysis



# Heat Integration: Stream Definition and Pinch Analysis by TI Method
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2017
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

from __future__ import annotations

import pandas as pd
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .stream import Stream  # type checking: avoids circular imports


@dataclass
class HENDesign:
    """
    Container for HEN Design.

    Examples:

    Args:

    Returns:

    """

    n_stages: int  # 0: heaters, 1-k, process stream HX, k+1: coolers
    streams: list[Stream] = field(default_factory=list)
    # Q exchanged between hot stream i and cold stream j in stage k
    Q_exchange: Optional[pd.DataFrame] = None
    # Hot stream temperatures while entering each stage
    T_hot_streams: Optional[pd.DataFrame] = None
    # Cold stream temperatures while leaving each stage
    T_cold_streams: Optional[pd.DataFrame] = None



# Ternary VLE Analysis
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2018
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

import numpy as np
import scipy as sci
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import warnings


class TernarySystem:
    """
    Container for a ternary mixture.

    Notes:
        This class supports a homogeneous, non-ideal liquid phase described by
            the Wilson activity coefficient model (Wilson 1964), in equilibrium
            with an ideal vapor phase (Wilson GM, J. Am. Chem. Soc. 86: 127-
            130, 1964). Vapor pressures are described by the Antoine equation
            in NIST format: log10(P) = A - (B / (T + C), with P in bar and
            T in K.

    Examples:
        sys567 = TernarySystem(  # pentane/hexane/heptane (ideal)
            component_ids=('pen', 'hex', 'hep'),
            component_names={'n-pentane',
                             'n-hexane',
                             'n-heptane'},
            axes={'pen': 0, 'hep': 1},
            Antoine={'pen': {'A': 3.989,   'B': 1070.617, 'C': -40.454},
                     'hex': {'A': 4.00266, 'B': 1171.53,  'C': -48.784},
                     'hep': {'A': 4.02832, 'B': 1268.636, 'C': -56.199}},
                     Wilson_a={'pen': {'pen': 0, 'hex': 0, 'hep': 0},
                               'hex': {'pen': 0, 'hex': 0, 'hep': 0},
                               'hep': {'pen': 0, 'hex': 0, 'hep': 0}},
                     Wilson_b={'pen': {'pen': 0, 'hex': 0, 'hep': 0},
                               'hex': {'pen': 0, 'hex': 0, 'hep': 0},
                               'hep': {'pen': 0, 'hex': 0, 'hep': 0}}
        )

        sysACM = TernarySystem(  # acetone/methanol/chloroform (non-ideal)
            component_ids={'acetone', 'methanol', 'chloroform'},
            component_names={'acetone', 'methanol', 'chloroform'},
            axes={'acetone': 0, 'methanol': 1},
            Antoine={'acetone':    {'A': 4.4245, 'B': 1312.3, 'C': -32.445},
                     'methanol':   {'A': 5.2041, 'B': 1581.3, 'C': -33.5},
                     'chloroform': {'A': 4.2077, 'B': 1233.1, 'C': -40.9530}},
            Wilson_a={
                'acetone': {
                    'acetone': 0,
                    'methanol': 0,
                    'chloroform': -0.7683
                    },
                'methanol': {
                    'acetone': 0,
                    'methanol': 0,
                    'chloroform': 0
                    },
                'chloroform': {
                    'acetone': -0.7191,
                    'methanol': 0,
                    'chloroform': 0
                    }
                },
            Wilson_b={
                'acetone': {
                    'acetone': 0,
                    'methanol': -115.663,
                    'chloroform': 262.1790
                    },
                'methanol': {
                    'acetone': -108.5260,
                    'methanol':  0,
                    'chloroform': -652.8960
                    },
                'chloroform': {
                    'acetone': 435.1440,
                    'methanol': -32.5972,
                    'chloroform': 0
                    }
                },
            P = 1.013
    )

    Args:
        component_ids: Short ids (abbreviations or formulas) for components.
            Enclose component_ids in parentheses.
        component_names: Names or full names of components.
        axes: Dictionary indicating which component is represented on which
            axis. In the example above, component 'pen' is represented on axis
            0 (x-axis) and component 'hep' is represented on axis 1 (y-axis).
        Antoine:  Dictionary of Antoine equation constants from NIST WebBook
            in NIST format (P in bar, T in K).
        Wilson_a, Wilson_b: a_ij and b_ij parameters for Wilson activity
            coefficient correlation.
            - Diagonal elements a_ii and b_ii are always zero.
            - For an ideal liquid mixture, set all a_ij, b_ij to zero.
            - For non-ideal mixtures, obtain a_ij and b_ij from the literature
                or Aspen.
            - To get a_ij and b_ij from Aspen:
             - - Create an Aspen file (in ``apwz`` or ``apw`` format)
                 containing the three compounds in the mixture.
             - - Go to Methods on the left menu pane and select WILSON as the
                 method with default settings.
             - - Navigate to Methods->Parameters->Binary Interaction->Wilson.
             - - Read the parameters a_ij and b_ij from the displayed table.
        P: Pressure of the system (bar). Defaults to 1.013 bar.

    Returns:
        A TernarySystem object containing the specified components, Antoine
            parameters and Wilson parameters.
    """

    def __init__(self, component_ids, component_names, axes,
                 Antoine, Wilson_a, Wilson_b, P=1.013, Tmin=None, Tmax=None):
        self.component_ids = component_ids
        self.components = self.component_ids
        self.component_names = component_names
        self.names = self.component_names
        self.axes = axes
        self.Antoine = Antoine
        self.Wilson_a = Wilson_a
        self.Wilson_b = Wilson_b
        self.P = P

        if self.axes is None:
            raise ValueError('Axes need to be specified')
        if 0 not in list(self.axes.values()):
            raise KeyError('Axis #0 (x-axis) must be specified')
        if 1 not in list(self.axes.values()):
            raise KeyError('Axis #1 (y-axis) must be specified')
        for j in self.components:  # identify implicit component if unspecified
            if j not in self.axes:
                axes[j] = 2

        Ts = self.Tsat(P)

        if Tmin is None:
            Tmin = min(Ts)  # minimum boiling point of system if ideal
        self.Tmin = Tmin

        if Tmax is None:
            Tmax = max(Ts)  # maximum boiling point of system if ideal
        self.Tmax = Tmax

        Tr = (Tmin + Tmax) / 2  # representative temperature of the system
        self.Tr = Tr

    def __repr__(self):
        rows_Ant = '\n'.join(f'          {comp}: {row}'
                             for comp, row in self.Antoine.items())
        rows_a = '\n'.join(f'        {comp}: {row}'
                           for comp, row in self.Wilson_a.items())
        rows_b = '\n'.join(f'        {comp}: {row}'
                           for comp, row in self.Wilson_b.items())
        disp_text = ('Ternary System:\n'
                     f'    component ids:\n\t{self.components}\n'
                     f'    component names:\n\t{self.component_names}\n'
                     f'    axes:\n\t{self.axes}\n'
                     f'    Antoine:\n{rows_Ant}\n'
                     f'    Wilson_b:\n{rows_a}\n'
                     f'    Wilson_b:\n{rows_b}\n')
        return disp_text

    # Temperatures
    T0 = 273.15  # 0 °C in K

    def Psat(self, T):
        """
        Calculate vapor pressure (bar) as a function of T (°C).

        Examples:
            sys567.Psat(60)
            sys567.Psat([60, 100])

        Args:
            T: Temperature in °C. T can be a scalar or an array.

        Returns:
            An array of vapor pressure(s) ``Psat`` at the specified
                temperature(s), with components ordered according to the
                ``axes`` specification used when constructing the
                ``TernarySystem``.
        """

        T = np.asarray(T)
        Psat = np.zeros(T.shape + (3,), dtype=float)  # to vectorize
        for j in self.components:
            if 'A' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter A is missing for {j}")
            if 'B' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter B is missing for {j}")
            if 'C' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter C is missing for {j}")

            A = self.Antoine[j]['A']
            B = self.Antoine[j]['B']
            C = self.Antoine[j]['C']

            Ps = 10**(A - B / (T + self.T0 + C))
            Psat[..., self.axes[j]] = Ps
        return Psat

    def Tsat(self, P):
        """
        Calculate boiling point (°C) as a function of P (bar).

        Examples:
            sys567.Tsat(1.013)
            sys567.Tsat([0.1, 1.0])

        Args:
            P: Pressure in bar. P can be a scalar or an array.

        Returns:
            An array of boiling point(s) ``Tsat`` at the specified
                pressure(s), with components ordered according to the
                ``axes`` specification used when constructing the
                ``TernarySystem``.
        """

        P = np.asarray(P)
        Tsat = np.zeros(P.shape + (3,), dtype=float)  # to vectorize
        for j in self.components:
            if 'A' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter A is missing for {j}")
            if 'B' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter B is missing for {j}")
            if 'C' not in self.Antoine[j]:
                raise KeyError(f"Antoine parameter C is missing for {j}")

            A = self.Antoine[j]['A']
            B = self.Antoine[j]['B']
            C = self.Antoine[j]['C']

            Ts = B / (A - np.log10(P)) - C - self.T0
            Tsat[..., self.axes[j]] = Ts
        return Tsat

    def gamma(self, x, T):
        """
        Calculate Wilson activity coefficient as a function of x and T (°C).

        Examples:
            sys567.gamma([0.32, 0.40, 0.28], 60)
            sys567.gamma([[0.32, 0.40, 0.28], [0.46, 0.18, 0.36]], [60, 100])

        Args:
            x: An array of mole fractions, with components ordered
                according to the ``axes`` specification used when
                constructing the ``TernarySystem``. All three mole fractions
                must be specified and must add up to 1. This argument can
                also contain multiple composition points in the form of an
                n x 3 matrix or an m x n x 3 meshgrid, as long as the last
                dimension contains the three mole fractions.
            T: Temperature in °C. If specifying a single composition, T
                should be scalar. If specifying multiple compositions, T
                should be an array with exactly as many elements as the
                number of compositions specified.
        Returns:
            An activity coefficient array ``gamma_i`` of dimension ... x 3,
                with components ordered according to the ``axes``
                specification used when constructing the ``TernarySystem``.
        """

        T = np.asarray(T)
        A = np.zeros(T.shape + (3, 3), dtype=float)  # to vectorize
        for i in self.components:
            for j in self.components:
                A[..., self.axes[i], self.axes[j]] = np.exp(
                    self.Wilson_a[i][j] + self.Wilson_b[i][j] / (T + self.T0)
                )

        x = np.asarray(x)
        Ax = np.einsum('...ij,...j->...i', A, x)
        with np.errstate(divide='ignore', invalid='ignore'):
            xAx = x / Ax
            lng = 1 - np.log(Ax) - np.einsum('...ji,...j->...i', A, xAx)
            g = np.exp(lng)
            return g

    def delxt(self, t, x):
        """
        Calculate the derivative of x with respect to t and express it as
            a function of (t, x).

        Note:
            This function returns the derivative of x without the equilibrium
                temperature. This derivative is used in the computation of
                residue curves.

        Example:
            sys567.delxt(1, [0.32, 0.40, 0.28])

        Args:
            t: A dimensionless, timelike variable.
            x: An array of mole fractions, with components ordered
                according to the ``axes`` specification used when
                constructing the ``TernarySystem``. All three mole fractions
                must be specified and must add up to 1. This argument can
                also contain multiple composition points in the form of an
                n x 3 matrix or an m x n x 3 meshgrid, as long as the last
                dimension contains the mole fractions.

        Returns:
            A derivative array ``dx/dt`` of dimension ... x 3, with
                components ordered according to the ``axes`` specification
                used when constructing the ``TernarySystem``.
        """

        x = np.asarray(x)
        x_shape = x.shape[:-1]
        Tr = np.full(x_shape, [self.Tr])
        T = np.empty(x_shape, dtype=float)
        delxt = np.empty_like(x, dtype=float)

        for i in np.ndindex(x_shape):
            xi = x[i]
            Tr0 = Tr[i]

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                res = sci.optimize.root_scalar(
                    lambda T: 1 - np.sum(xi * self.gamma(xi, T) *
                                         self.Psat(T) / self.P),
                    x0=Tr0)
            Ti = res.root
            T[i] = Ti
            g = self.gamma(xi, T)
            Ps = self.Psat(Ti)
            delxt[i] = xi - xi * g * Ps / self.P
        return delxt

    def delxT(self, x):
        """
        Calculate the derivative of x with respect to t, express it as a
            function of x and return it with the corresponding equilibrium
            temperature.

        Note:
            This function returns the derivative of x and the corresponding
                equilibrium temperature, i.e., the bubble point or dew point.
                These values are used in the plotting of equilibrium fields
                as well as in flash and distillation calculations.

        Examples:
            sys567.delxT([0.32, 0.40, 0.28])
            delx, T = sys567.delxT([0.32, 0.40, 0.28])

        Args:
            x: An array of mole fractions, with components ordered
                according to the ``axes`` specification used when
                constructing the ``TernarySystem``. All three mole fractions
                must be specified and must add up to 1. This argument can
                also contain multiple composition points in the form of an
                n x 3 matrix or an m x n x 3 meshgrid, as long as the last
                dimension contains the mole fractions.

        Returns:
            A derivative array ``dx/dt`` of dimension ... x 3, with
                components ordered according to the ``axes`` specification
                used when constructing the ``TernarySystem``.
            An equilibrium temperature value or an array of equilibrium
                temperatures corresponding to the derivatives.
        """

        x = np.asarray(x)
        x_shape = x.shape[:-1]
        Tr = np.full(x_shape, [self.Tr])
        T = np.empty(x_shape, dtype=float)
        delxT = np.empty_like(x, dtype=float)

        for i in np.ndindex(x_shape):
            xi = x[i]
            Tr0 = Tr[i]

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                res = sci.optimize.root_scalar(
                    lambda T: 1 - np.sum(xi * self.gamma(xi, T) *
                                         self.Psat(T) / self.P),
                    x0=Tr0
                )
            Ti = res.root
            T[i] = Ti

            g = self.gamma(xi, Ti)
            Ps = self.Psat(Ti)
            delxT[i] = xi - xi * g * Ps / self.P
        return delxT, T

    def delx(self, x):
        """
        Calculate the derivative of x with respect to t and express it
            as a function of x.

        Note:
            This function returns the derivative of x without the equilibrium
                temperature.

        Example:
            sys567.delx([0.32, 0.40, 0.28])

        Args:
            x: An array of mole fractions, with components ordered
                according to the ``axes`` specification used when
                constructing the ``TernarySystem``. All three mole fractions
                must be specified and must add up to 1. This argument can
                also contain multiple composition points in the form of an
                n x 3 matrix or an m x n x 3 meshgrid, as long as the last
                dimension contains the mole fractions.

        Returns:
            A derivative array ``dx/dt`` of dimension ... x 3, with
                components ordered according to the ``axes`` specification
                used when constructing the ``TernarySystem``.
        """

        x = np.asarray(x)
        x_shape = x.shape[:-1]
        Tr = np.full(x_shape, [self.Tr])
        T = np.empty(x_shape, dtype=float)
        delx = np.empty_like(x, dtype=float)

        for i in np.ndindex(x_shape):
            xi = x[i]
            Tr0 = Tr[i]

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                res = sci.optimize.root_scalar(
                    lambda T: 1 - np.sum(xi * self.gamma(xi, T) *
                                         self.Psat(T) / self.P),
                    x0=Tr0
                )
            Ti = res.root
            T[i] = Ti

            g = self.gamma(xi, Ti)
            Ps = self.Psat(Ti)
            delx[i] = xi - xi * g * Ps / self.P
        return delx

    def delx2(self, x2):
        """
        Calculate the 2-element derivative of a 2-element vector x with
            respect to t and express it as a function of x. The elements of
            the vector and the derivative correspond with the principal
            axes, i.e., axis 0 (horizontal axis) and axis 1 (vertical axis).

        Note: This function returns the derivative of x without the equilibrium
                temperature. The derivative is used in computing the Jacobian
                and determining stationary points.

        Example:
            sys567.delx([0.32, 0.40])

        Args:
            x2: An array of two mole fractions, with components listed in
                the order specified by the axes argument of the system spec.
                Only two mole fractions must be specified, and they need not
                add up to 1. This argument can also contain multiple
                composition points in the form of an n x 2 matrix or an
                m x n x 2 meshgrid, as long as the last dimension contains
                the two mole fractions.

        Returns:
            A derivative array ``dx/dt`` of dimension ... x 2, with
                the two components ordered according to the ``axes``
                specification used when constructing the ``TernarySystem``.
        """

        x2 = np.asarray(x2)
        if x2.ndim == 1:
            x3 = 1.0 - x2.sum()
            x_full = np.array([x2[0], x2[1], x3])
        else:
            if x2.shape[-1] == 2:
                x_last = x2
            elif x2.shape[0] == 2:
                x_last = np.moveaxis(x2, 0, -1)
            x3 = 1.0 - x_last.sum(axis=-1, keepdims=True)
            x_full = np.concatenate([x_last, x3], axis=-1)

        delx = self.delx(x_full)
        return delx[..., :2]

    def delx2t(self, t, x2):
        """
        Calculate the 2-element derivative of a 2-element vector x2 with
            respect to t and express it as a function of (t, x2).

        Note:
            This function returns the derivative of x without the equilibrium
                temperature. This derivative is used in the computation of
                residue curves and separatrices.

        Example:
            sys567.delx2t(0, [0.32, 0.40])

        Args:
            t: A dimensionless, timelike variable.
            x2: An array of two mole fractions, with components listed in
                the order specified by the axes argument of the system spec.
                Only two mole fractions must be specified, and they need not
                add up to 1. This argument can also contain multiple
                composition points in the form of an n x 2 matrix or an
                m x n x 2 meshgrid, as long as the last dimension contains
                the two mole fractions.

        Returns:
            A derivative array ``dx/dt`` of dimension ... x 2, with
                the two components ordered according to the ``axes``
                specification used when constructing the ``TernarySystem``.
        """

        x2 = np.asarray(x2)
        if x2.ndim == 1:
            x3 = 1.0 - x2.sum()
            x_full = np.array([x2[0], x2[1], x3])
        else:
            if x2.shape[-1] == 2:
                x_last = x2
            elif x2.shape[0] == 2:
                x_last = np.moveaxis(x2, 0, -1)
            x3 = 1.0 - x_last.sum(axis=-1, keepdims=True)
            x_full = np.concatenate([x_last, x3], axis=-1)

        delx = self.delx(x_full)
        return delx[..., :2]

    def yeq(self, x):
        """
        Determine a saturated vapor composition y in equilibrium with a
            saturated liquid composition x, and the bubble point temperature.

        Example:
            sys567.yeq([0.32, 0.40, 0.28])

        Args:
            x: An array of mole fractions, with components listed in the
                order specified by the axes argument of the system spec.
                All three mole fractions must be specified and  must
                add up to 1. This argument can also contain multiple
                composition points in the form of an n x 2 matrix or an
                m x n x 2 meshgrid, as long as the last dimension contains
                the two mole fractions.

        Returns:
            An array of dimension ... x 3 containing the saturated vapor
                composition(s), with components ordered according to the
                ``axes`` specification used when constructing the
                ``TernarySystem``.
            A bubble point temperature value or an array of bubble point
                temperatures corresponding to the saturated vapor
                compositions.
        """

        dx, T = self.delxT(x)
        y = x - dx
        return y, T

    def xeq(self, y):
        """
        Determine a saturated liquid composition x in equilibrium with a
            saturated vapor composition y and the dew point temperature.

        Example:
            sys567.xeq([0.32, 0.40, 0.28])

        Args:
            x: An array of mole fractions, with components listed in the
                order specified by the axes argument of the system spec.
                All three mole fractions must be specified and must add
                up to 1. This argument can also contain multiple
                composition points in the form of an n x 2 matrix or an
                m x n x 2 meshgrid, as long as the last dimension contains
                the two mole fractions.

        Returns:
            An array of dimension ... x 3 containing the saturated liquid
                composition(s), with components ordered according to the
                ``axes`` specification used when constructing the
                ``TernarySystem``.
            A dew point temperature value or an array of dew point
                temperatures corresponding to the saturated liquid
                compositions.

        """

        y = np.asarray(y)
        y_shape = y.shape[:-1]
        x = np.empty_like(y)
        T = np.empty(y_shape)

        for idx in np.ndindex(y_shape):
            yi = y[idx]

            X0 = np.concatenate((yi, [self.Tr]))

            res = sci.optimize.root(
                lambda X: np.append(yi * self.P - X[:3] *
                                    self.gamma(X[:3], X[3]) *
                                    self.Psat(X[3]),
                                    np.sum(X[:3]) - 1),
                x0=X0)
            x[idx] = res.x[:3]
            T[idx] = res.x[3]
        return x, T

    def plot_vle_field(self,
                       fsize=8, font=12, facecolor='white',
                       xy_tol=-1e-6, n_vectors=31, arrow_scale=31,
                       Tmin=None, Tmax=None, contour_T_heavy=10,
                       contour_T_medium=5, contour_T_light=1,
                       **kwargs):
        """
        Plot equilibrium field and constant-temperature contours for a
            ternary vapor-liquid mixture in [x1,x2] or [y1,y2] space.

        Examples:
            fig, ax, _ = sys567.plot_vle_field(
                n_vectors=51, arrow_scale=51,
                contour_T_heavy=10, contour_T_medium=5, contour_T_light=1)

            fig, ax, quiver = sysACM.simulate_vle_field(
                n_vectors=51, arrow_scale=51,
                Tmin=50, Tmax=70,
                contour_T_heavy=1, contour_T_medium=0.5, contour_T_light=0.1)

        Args:
            fsize: Figure size. A square figure with this length is created.
                Defaults to 8.
            font: Base font used for axis labels, etc. Defaults to 12.
            xy_tol: Tolerance for right triangle. Set to slightly less than
                zero, e.g., -1e-6. Defaults to -1e-6.
            n_vectors: Number of e-vectors along each axis. Defaults to 31.
            arrow_scale: Scale of arrows on equilibrium field. Set it
                approximately equal to ``n_vector``. Defaults to 31.
            T_min, T_max: Minimum and maximum temperatures of the system. To
                get all temperature contours to plot, these should be set to
                the minimum and maximum boiling points in the system.
            contour_heavy_T: Temperature increment (°C) for heavy, labeled
                temperature contours. Defaults to 10 °C.
            contour_medium_T: Temperature increment (°C) for medium, unlabeled
                temperature contours. Defaults to 5 °C.
            contour_light_T: Temperature increment (°C) for light, unlabeled
                temperature contours. Defaults to 1 °C.
            **kwargs: Additional keyword arguments passed to
                matplotlib.pyplot.quiver. Common options include ``color``
                and ``linewidth``.
        Returns:
            Figure: Figure object containing the equilibrium field plot.
            Axes: Associated axes object.
        """

        components = self.components
        axes = self.axes
        P = self.P

        def Psat(T):
            return self.Psat(T)

        def gamma(x, T):
            return self.gamma(x, T)

        def delxT(x):
            return self.delxT(x)

        def yeq(x):
            return self.yeq(x)

        def xeq(y):
            return self.xeq(y)

        if Tmin is None:
            Tmin = self.Tmin
        if Tmax is None:
            Tmax = self.Tmax
        Tr = (Tmin + Tmax) / 2  # representative temperature of the system

        fig, ax = plt.subplots(figsize=(fsize, fsize), facecolor=facecolor)
        ax.set_aspect('equal', adjustable='box')
        for spine in ax.spines.values():
            spine.set_visible(True)
            spine.set_linewidth(1)
        ax.set_xlim([0, 1])
        ax.set_ylim([0, 1])
        ax.grid(True)
        for j in components:
            if axes[j] == 0:
                ax.set_xlabel(j, fontsize=font)
            if axes[j] == 1:
                ax.set_ylabel(j, fontsize=font)

        X = np.linspace(0, 1, n_vectors)
        xx, yy = np.meshgrid(X, X)

        XY_full = np.stack([xx, yy, 1 - xx - yy], axis=-1)
        mask = 1 - xx - yy >= xy_tol
        XY_flat = XY_full[mask]

        uv_flat, _ = delxT(XY_flat)

        uv = np.full(xx.shape + (3,), np.nan)
        uv[mask] = uv_flat
        uu = uv[..., 0]
        vv = uv[..., 1]

        # Normalize
        mag = np.hypot(uu, vv)
        un, vn = uu, vv
        with np.errstate(invalid='ignore'):
            un = np.where(np.isnan(un), np.nan,
                          np.where(mag > 0, np.divide(uu, mag), 0))
            vn = np.where(np.isnan(vn), np.nan,
                          np.where(mag > 0, np.divide(vv, mag), 0))

        quiver = ax.quiver(X, X, un, vn, angles='xy',
                           color='#8888ff', scale=arrow_scale,
                           alpha=0.5, **kwargs)

        # Calculate and plot temperature contours
        x_contours = 51
        XT = np.linspace(0, 1, x_contours)
        X0, X1 = np.meshgrid(XT, XT)
        Y0, Y1 = np.meshgrid(XT, XT)
        TT0, TT1 = np.meshgrid(XT, XT)

        for i in range(np.size(XT)):
            for j in range(np.size(XT)):
                if XT[i] + XT[j] <= 1:
                    x = [XT[i], XT[j], 1 - XT[i] - XT[j]]
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore', RuntimeWarning)
                        res = sci.optimize.root_scalar(
                            lambda T:
                            1 - np.sum(x * gamma(x, T) * Psat(T) / P),
                            x0=Tr
                        )
                    T = res.root

                    X0[i, j] = XT[i]
                    X1[i, j] = XT[j]
                    TT0[i, j] = T

                    y = x * gamma(x, T) * Psat(T) / P
                    Y0[i, j] = y[0]
                    Y1[i, j] = y[1]
                    TT1[i, j] = T
                else:
                    TT0[i, j] = np.nan
                    TT1[i, j] = np.nan

        def clevels(c):
            tmin = np.round(Tmin / c) * c
            tmax = np.round(Tmax / c + 1) * c
            clevels = np.arange(tmin, tmax, c)
            return clevels

        cl = ax.contour(X0, X1, TT0,
                        levels=clevels(contour_T_heavy),
                        linewidths=0.5, colors='#0000ff')
        ax.clabel(cl, inline=True, fontsize=8)
        ax.contour(X0, X1, TT0,
                   levels=clevels(contour_T_medium),
                   linewidths=0.25, colors='#0000ff')
        ax.contour(X0, X1, TT0,
                   levels=clevels(contour_T_light),
                   linewidths=0.05, colors='#0000ff')

        cl = ax.contour(Y0, Y1, TT1,
                        levels=clevels(contour_T_heavy),
                        linewidths=0.5, colors='#ff0000')
        ax.clabel(cl, inline=True, fontsize=8)
        ax.contour(Y0, Y1, TT1,
                   levels=clevels(contour_T_medium),
                   linewidths=0.25, colors='#ff0000')
        ax.contour(Y0, Y1, TT1,
                   levels=clevels(contour_T_light),
                   linewidths=0.05, colors='#ff0000')

        # Wrap up, save and display plot
        ax.plot([0, 1], [0, 0], 'k-')  # right triangle, horizontal line
        ax.plot([0, 0], [0, 1], 'k-')  # right triangle, vertical line
        ax.plot([0, 1], [1, 0], 'k-')  # right triangle, diagonal line

        plt.draw()
        return fig, ax, quiver

    draw_vle_field = plot_vle_field
    simulate_vle_field = plot_vle_field

    def plot_residue_curve(self, ax, feed=None,
                           integration_direction='both',
                           integration_time=10, max_step=np.inf,
                           show_feed=True,
                           separatrix=False, **kwargs):
        """
        Plot (a) residue curve(s) for a ternary vapor-liquid mixture from a
            specified (a) feed point(s).

        Examples:
            sys567.plot_residue_curve(
                ax, feed=[
                    [0.33, 0.33, 0.34],
                    [0.32, 0.40, 0.28]
                ],
                show_feed=True, max_step=0.1, color='k', ls='-', lw=1.5)

            sys567.plot_residue_curve(
                ax, feed={
                    0: {'pen': 0.33, 'hex': 0.34, 'hep': 0.33},
                    1: {'pen': 0.32, 'hex': 0.28, 'hep': 0.40}
                    },
                    show_feed=True, max_step=0.1, color='k', ls='-', lw=1.5
            )

            ax, res_fwd, res_bwd = sysACM.plot_residue(
                ax, feed={
                    0: {'acetone': 0.4, 'methanol': 0.2, 'chloroform': 0.4},
                    1: {'acetone': 0.6, 'methanol': 0.2, 'chloroform': 0.2},
                    2: {'acetone': 0.2, 'methanol': 0.6, 'chloroform': 0.2},
                    3: {'acetone': 0.5, 'methanol': 0.4, 'chloroform': 0.1}
                },
                integration_time=20, max_step=0.5, show_feed=True,
                linestyle='--', color='k', marker='o', markersize=6
            )

        Args:
            ax: Axes on which residue curve(s) is (are) to be plotted.
            feed: Feed point. One or more feeds can be specified as a
                1 x 3 vector or an n x 3 matrix, as long as the last
                dimension contains the three mole fractions with components
                ordered according to the ``axes`` specification used when
                constructing the ``TernarySystem``. The feed(s) can also be
                as a nested dictionary as shown in the example above.
            integration_direction: Direction of integration. Options are:
                ``both``: forward and backward in time,
                ``positive``, ``pos``, ``forward``, ``fwd``: forward in time,
                ``negative``, ``neg``, ``backward``, ``bwd``: backward in time.
                Defaults to 'both'.
            integration_time: Integration time for forward and backward residue
                curves. Defaults to 10. If different integration times are
                needed in each direction, specify the curves separately setting
                ``integration_direction`` to ``pos`` and ``neg``, respectively.
            max_step: Maximum integration step. Defaults to infinity. Set it
                to a low value, e.g., 0.05 or 0.10, if the residue curve
                is not continuous or contains kinks. Setting this
                parameter to < 1 will slow down residue curve plotting.
                Defaults to infinity (``np.inf``) for regular residue curves
                and 0.2 for separatrices.
            show_feed: True/False value indicating whether feed should be
                displayed as a point.
            separatrix: True/False value indicating whether residue curve is
                a separatrix. This is automatically set to True when called
                by ``plot_separatrices``.
            **kwargs: Additional keyword arguments passed to matplotlib.pyplot.
                Common options include ``linewidth`` for the residue curve and
                ``marker`` for the feed point(s).

        Returns:
            Axes passed to the function and if requested, objects containing
                the residue curve coordinates and other results from
                ``scipy.optimize.solve_ivp``.
        """

        # Parse kwargs
        common_keys = {'color', 'c',
                       'alpha'}  # shared between point and line

        point_keys = {
                'marker',
                'markersize', 'ms',
                'markerfacecolor', 'mfc',
                'markeredgecolor', 'mec',
                'markeredgewidth', 'mew',
                'fillstyle',
            }

        line_keys = {
            'linewidth', 'lw',
            'drawstyle',
            'dash_capstyle',
            'dash_joinstyle',
            'solid_capstyle',
            'solid_joinstyle',
            'antialiased',
        }

        common_kwargs, point_kwargs, line_kwargs = {}, {}, {}

        for key, value in kwargs.items():
            if key in common_keys:
                common_kwargs[key] = value
            elif key in point_keys:
                point_kwargs[key] = value
            elif key in line_keys:
                line_kwargs[key] = value
            else:
                pass

        point_kwargs = {**common_kwargs, **point_kwargs}
        line_kwargs = {**common_kwargs, **line_kwargs}

        if feed is None:
            raise ValueError(
                'Feed point(s) must be specified ' +
                'if residue curve is requested'
            )

        if not separatrix:
            # If not plotting a separatrix, a 3-D derivative is being used
            # Check the dimension and mole fraction sum of the feed
            if isinstance(feed, dict):
                row_sums = {i: sum(d.values()) for i, d in feed.items()}
                for i, total in row_sums.items():
                    if abs(total - 1) > 1e-6:
                        raise ValueError(
                            'Residue curve feed mole fractions must add ' +
                            f"up to 1: got {total} for row {i}"
                        )
            elif isinstance(feed, (list, np.ndarray)):
                feed = np.asarray(feed)
                if feed.ndim == 1:  # feed is a vector: single feed point
                    total = np.sum(feed)
                    if abs(total - 1) > 1e-6:
                        raise ValueError(
                            'Residue curve feed mole fractions must add ' +
                            f"up to 1: got {total}"
                        )
                elif feed.ndim == 2:  # feed is a matrix: multiple feed points
                    total = np.sum(feed, axis=-1)
                    for i in range(len(total)):
                        if abs(total[i] - 1) > 1e-6:
                            raise ValueError(
                                'Residue curve feed mole fractions must add \n'
                                + f"up to 1: got {total[i]} for row {i}"
                            )
                else:
                    raise ValueError(
                        'Residue curve feed must have dimension <= 2'
                    )
            else:
                pass  # feed is of unexpected data type

        if isinstance(feed, dict):
            x_feed = np.zeros((len(feed), 3))
            ii = 0
            for i in feed:
                for jj in feed[i]:
                    x_feed[ii, self.axes[jj]] = feed[i][jj]
                ii += 1
        else:
            x_feed = feed

        if not separatrix:
            func = self.delxt  # 3-D derivative
            events = out_of_triangle  # 3-D version
        else:
            func = self.delx2t  # 2-D derivative
            events = out_of_triangle2  # 2-D version

        res_fwd = []
        res_bwd = []
        for row in np.atleast_2d(x_feed):
            xF = row
            if show_feed:
                ax.plot(xF[0], xF[1], **point_kwargs)
            with np.errstate(all='ignore'):
                int_t = integration_time
                if integration_direction.lower() in ('both',
                                                     'positive',
                                                     'forward',
                                                     'pos', 'fwd'):
                    residue_fwd = sci.integrate.solve_ivp(func,
                                                          [0, int_t], xF,
                                                          method='DOP853',
                                                          events=events,
                                                          max_step=max_step)
                    ax.plot(residue_fwd.y[0, :], residue_fwd.y[1, :],
                            **line_kwargs)
                else:
                    residue_fwd = []
                res_fwd.append(residue_fwd)

                if integration_direction.lower() in ('both',
                                                     'negative',
                                                     'backward',
                                                     'neg', 'bwd'):
                    residue_bwd = sci.integrate.solve_ivp(func,
                                                          [0, -int_t], xF,
                                                          method='DOP853',
                                                          events=events,
                                                          max_step=max_step)
                    ax.plot(residue_bwd.y[0, :], residue_bwd.y[1, :],
                            **line_kwargs)
                else:
                    residue_bwd = []
                res_bwd.append(residue_bwd)

        return ax, res_fwd, res_bwd

    plot_residue = plot_residue_curve
    draw_residue = draw_residue_curve = plot_residue_curve
    simulate_residue = simulate_residue_curve = plot_residue_curve
    integrate_residue = integrate_residue_curve = plot_residue_curve

    def vle_properties(self, z, decimals=3):
        """
        Calculate VLE properties of a ternary mixture defined by a composition
            vector z.

        Example:
            system.vle_properties([0.3, 0.6, 0.1])
            system.mixture_properties([[0.3, 0.6, 0.1], [0.2, 0.3, 0.5]])

        Args:
            z: The overall composition of the mixture, averaged over the
                liquid and vapor phases. The coordinates should be ordered
                according to the ``axes`` specification used when
                constructing the ``TernarySystem``. All three mole fractions
                must be specified and must add up to 1. This argument can
                also contain multiple composition points in the form of an
                n x 3 matrix or an m x n x 3 meshgrid, as long as the last
                dimension contains the three mole fractions.

        Returns:
            A VLEProperties object containing the compositions and temperatures
                of the mixture's bubble point and dew point.
        """
        ye, Tb = self.yeq(z)
        xe, Td = self.xeq(z)
        p = VLEProperties(P=self.P,
                          molefrac=z,
                          yeq=ye,
                          xeq=xe,
                          T_bubl=Tb,
                          T_dew=Td,
                          decimals=decimals)
        return p

    mixture_properties = vle_properties

    def find_stationary_points(self,
                               n_complex=500, iters=1,
                               tol=1e-6, ftol=1e-12,
                               decimals=3):
        """
        Find the stationary points of a ternary vapor-liquid mixture, and
            determine their stability by computing the eigenvalues of the
            system Jacobian. Stationary points are compositions at which
            ``y = x`` and ``dx/dt = 0``.

        Examples:
            sys567.find_stationary_points()
            sysACM.find_stationary_points()

        Args:
            n_complex: Number of points in the simplicial complex in the
                search algorithm. Defaults to 500.
            iters: Number of iterations in the construction of the simplicial
                complex. Defaults to 1.
            ftol: Tolerance for stationary points. Defaults to 1e-12.
            tol: Tolerance for checking positivity or negativity of
                eigenvalues to determine stability. Defaults to 1e-6.
            decimals: Number of decimals used in displaying output.
                Defaults to 3.

        Returns:
            A StationaryPoints object containing the coordinates of the
                stationary points, along with their boiling points (°C),
                eigenvalues (of the local Jacobian) and stabilities.

        """

        res = sci.optimize.shgo(
            lambda x: np.linalg.norm(self.delx(x)),
            bounds=[(0, 1), (0, 1), (0, 1)],
            constraints={sci.optimize.LinearConstraint(A=[1, 1, 1], ub=1)},
            n=n_complex, iters=iters, sampling_method='sobol',
            options={'ftol': ftol}
        )
        x_stat = np.vstack((res.xl, [[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
        x_stat = np.unique(np.round(x_stat, 6), axis=0)

        _, T = self.delxT(x_stat)

        x2 = x_stat[:, :2]
        eigenvals = np.empty((0, 2))
        stability = []
        for row in x2:
            A2 = sci.optimize.approx_fprime(row, self.delx2)
            e = sci.linalg.eigvals(A2).real
            eigenvals = np.vstack([eigenvals, e.ravel()])
            sign0 = sign_tol(e[0])
            sign1 = sign_tol(e[1])
            if sign0 == 1 and sign1 == 1:
                stability.append('unstable')
            elif sign0 == -1 and sign1 == -1:
                stability.append('stable')
            elif sign0 == -1 and sign1 == 1:
                stability.append('saddle')
            elif sign0 == 1 and sign1 == -1:
                stability.append('saddle')
            else:
                # Zero eigenvalue encountered
                # Linear system cannot be used to conclude stability
                stability.append('inconclusive')

        res = StationaryPoints(x=x_stat,
                               T=T,
                               eigenvals=eigenvals,
                               stability=stability,
                               decimals=decimals)
        return res

    def plot_separatrices(self, ax,
                          stationary_points=None,
                          vertex_tol=1e-5,
                          eigen_tol=1e-5,
                          separatrix_feed_dx=1e-6,
                          integration_time=100,
                          max_step=0.2,
                          **kwargs):
        """
        Plot separatrices for a ternary vapor-liquid mixture.

        Note:
            Separatrices are plotted from interior saddle points, i.e.,
                saddle points that are not vertices of the triangle.
                If no separatrices are found, the function returns either
                nothing or a list of stationary points.

        Examples:
            sysACM.plot_separatrices(ax, c='g', ls='-')
            sysACM.plot_separatrices(ax, stationary_points=res,
                                     c='g', ls='-')

        Args:
            ax: Axes on which separatrices are to be plotted.
            stationary_points: A StationaryPoints object containing the
                coordinates of the stationary points, their boiling points,
                eigenvalues and stabilities. This argument is optional. Specify
                if stationary points were previously computed. If not
                specified, stationary points will be computed before plotting
                separatrices.
            vertex_tol: Tolerance for checking whether a stationary point is
                on a vertex of the triangle. Defaults to 1e-5.
            eigen_tol: Tolerance for checking nonnegativity of an eigenvalue.
                Defaults to 1e-5.
            separatrix_feed_dx: The distance from an interior saddle point to
                the feed of each separatrix. Defaults to 1e-6.
            integration_time: Integration time for forward and backward residue
                branches of each separatrix. Defaults to 100.
            max_step: Maximum integration step. Defaults to infinity. Set it
                to a low value, e.g., 0.05 or 0.10, if the separatrix
                contains kinks or other artifacts of integration. Setting this
                parameter to < 1 will slow down separatrix plotting.
                Defaults to 0.2.
            **kwargs: Additional keyword arguments passed to matplotlib.pyplot.
                Common options include ``color`` (``c``), ``linewidth``
                (``lw``) and ``linestyle`` (``ls``).

        Returns:
            Axes passed to the function and if requested, objects containing
                the residue curve coordinates and other results from
                ``scipy.optimize.solve_ivp``.
        """

        if stationary_points is None:
            res = self.find_stationary_points()
        else:
            res = stationary_points

        residue = []
        for i in range(len(res.T)):
            if res.stability[i] == 'saddle':
                vertex_distance = euclidean_distance(
                    res.x[i], [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
                )  # distance from stationary point to all 3 vertices
                # If stationary point is not vertex
                if np.all(vertex_distance > vertex_tol):
                    x2 = res.x[i][:2]
                    A2 = sci.optimize.approx_fprime(x2, self.delx2)
                    w, v = sci.linalg.eig(A2)

                    for j in range(len(w.real)):
                        feed_a = x2 + v[j] * separatrix_feed_dx
                        feed_b = x2 - v[j] * separatrix_feed_dx

                        feed = []
                        feed.append(feed_a)
                        feed.append(feed_b)
                        feed = np.vstack(feed)

                        if w.real[j] > eigen_tol:
                            residue_pos = self.plot_residue_curve(
                                ax, feed=feed,
                                integration_direction='pos',
                                integration_time=integration_time,
                                max_step=max_step,
                                show_feed=False, separatrix=True,
                                **kwargs
                            )
                            residue.append(residue_pos)
                        elif w.real[j] < -eigen_tol:
                            residue_neg = self.plot_residue_curve(
                                ax, feed=feed,
                                integration_direction='neg',
                                integration_time=integration_time,
                                max_step=max_step,
                                show_feed=False, separatrix=True,
                                **kwargs
                            )
                            residue.append(residue_neg)
                        else:
                            pass  # eigenvalue is zero within tolerance
        return ax, residue

    plot_separatrix = plot_separatrices
    draw_separatrix = plot_separatrix
    simulate_separatrix = plot_separatrix
    integrate_separatrix = plot_separatrix
    draw_separatrices = plot_separatrices
    simulate_separatrices = plot_separatrices
    integrate_separatrices = plot_separatrices

    def atanm(self, x):
        """
        Modified arctan function, returns a result in the range (0, π).
        """

        ang = np.atan(x)
        if ang < 0:
            ang = ang + np.pi
        return ang

    def angle_dx(self, x):
        """
        Angle in the range (0, π) made by the equilibrium vector with the
            positive x-axis.
        """

        dx, T = self.delxT(x)
        if abs(dx[0]) <= 1e-6:
            angle = np.pi / 2
        else:
            slope = dx[1] / dx[0]
            angle = self.atanm(slope)
        return angle

    def angle_xy(self, x, y):
        """
        Angle in the range (0, π) made by a vector connecting the points
            x and y with the positive x-axis.
        """

        if np.abs(x[0] - y[0]) <= 1e-6:
            angle = np.pi / 2
        else:
            slope = (x[1] - y[1]) / (x[0] - y[0])
            angle = self.atanm(slope)
        return angle

    def flash_T_objective(self, x, xF, T):
        """
        Objective function for ``flash_at_T``.
        """

        z = ((self.angle_dx(x) - self.angle_xy(x, xF))**2 +
             (self.yeq(x)[1] - T)**2 +
             (self.xeq(self.yeq(x)[0])[1] - T)**2)
        return z

    def flash_at_T(self, xF, T=25, decimals=3):
        """
        Perform a flash calculation by specifying the flash temperature.
        """

        eq_cons = {'type': 'eq',
                   'fun': lambda x: np.sum(x) - 1}

        bounds = sci.optimize.Bounds([0, 0, 0], [1, 1, 1])
        res = sci.optimize.minimize(lambda x: self.flash_T_objective(x, xF, T),
                                    x0=xF,
                                    method='SLSQP',
                                    constraints=[eq_cons],
                                    bounds=bounds,
                                    tol=1e-9)

        x = res.x
        y = self.yeq(x)[0]
        T = self.yeq(x)[1]
        Vfrac = np.linalg.norm(x - xF) / np.linalg.norm(x - y)

        res = FlashResult(xF=xF,
                          x=x,
                          y=y,
                          T=T,
                          P=self.P,
                          Lfrac=1 - Vfrac,
                          Vfrac=Vfrac,
                          decimals=decimals)
        return res

    def flash_Vfrac_objective(self, x, xF, Vfrac=0.5):
        """
        Objective function for ``flash_at_Vfrac``.
        """

        x = np.array(x)
        Tx = self.vle_properties(x).T_bubl
        xF = np.array(xF)
        y, Ty = self.yeq(x)

        z = ((self.angle_dx(x) - self.angle_xy(x, xF))**2 +
             (np.linalg.norm(x - xF) / np.linalg.norm(x - y) - Vfrac)**2 +
             (Tx - Ty)**2)
        return z

    def flash_at_Vfrac(self, xF, Vfrac, decimals=3):
        """
        Perform a flash calculation by specifying the vapor fraction
            (``Vfrac``) of the flash evaporator.
        """

        eq_cons = {'type': 'eq',
                   'fun': lambda x: np.sum(x) - 1}

        bounds = sci.optimize.Bounds([0, 0, 0], [1, 1, 1])
        res = sci.optimize.minimize(
            lambda x: self.flash_Vfrac_objective(x, xF, Vfrac),
            x0=xF,
            method='SLSQP',
            constraints=[eq_cons],
            bounds=bounds,
            tol=1e-9)

        x = res.x
        y = self.yeq(x)[0]
        T = self.yeq(x)[1]
        Vfrac = np.linalg.norm(x - xF) / np.linalg.norm(x - y)

        res = FlashResult(xF=xF,
                          x=x,
                          y=y,
                          T=T,
                          P=self.P,
                          Lfrac=1 - Vfrac,
                          Vfrac=Vfrac,
                          decimals=decimals)
        return res

    def draw_flash_bfd(self, flash_res):
        """
        Draw a BFD of a flash evaporator.
        """

        res = flash_res

        P_str = np.round(res.P, 2)
        T_str = np.round(res.T, 2)
        xF_str = np.round(res.xF, 2)

        if res.x is None:
            x_str = '?'
        else:
            x_str = np.round(res.x, 2)

        if res.y is None:
            y_str = '?'
        else:
            y_str = np.round(res.y, 2)

        if res.Vfrac is None:
            Vfrac_str = '?'
        else:
            Vfrac_str = np.round(res.Vfrac, 2)

        fig1, ax1 = plt.subplots()
        plt.axis('off')
        _ = ax1.set_xlim(-4, 4)
        _ = ax1.set_ylim(-4, 4)

        patches.ArrowStyle.Curve

        feed = patches.Arrow(-3, 0, 2, 0, width=0.5, color='black')
        ax1.add_patch(feed)

        ax1.text(-1.1, 0.2,
                 f"xF = {xF_str}",
                 ha='right',
                 va='bottom',
                 fontsize=8,
                 color='black')

        vap = patches.Arrow(1, 1.6, 2, 0, width=0.5, color='red')
        ax1.add_patch(vap)

        ax1.text(1.1, 1.8,
                 f"y = {y_str}",
                 ha='left',
                 va='bottom',
                 fontsize=8,
                 color='red')

        liq = patches.Arrow(1, -1.6, 2, 0, width=0.5, color='blue')
        ax1.add_patch(liq)

        ax1.text(1.1, -1.4,
                 f"x = {x_str}",
                 ha='left',
                 va='bottom',
                 fontsize=8,
                 color='blue')

        unit = patches.Rectangle((-1, -2), 2, 4,
                                 linewidth=2,
                                 edgecolor='black',
                                 facecolor='white',
                                 alpha=1)
        ax1.add_patch(unit)

        ax1.text(0, 0.5, 'Flash',
                 ha='center',
                 va='center',
                 fontsize=10)

        ax1.text(0, 0, f"P = {P_str} bar",
                 ha='center',
                 va='center',
                 fontsize=8)

        ax1.text(0, -0.5, f"T = {T_str} °C",
                 ha='center',
                 va='center',
                 fontsize=8)

        ax1.text(2, 0, f"Vfrac = {Vfrac_str}",
                 ha='center',
                 va='center',
                 fontsize=8)

        return fig1, ax1


class VLEProperties:
    """
    Container for the properties of a ternary mixture at a specified
        composition.
    """

    def __init__(self, P, molefrac, yeq, xeq, T_bubl, T_dew, decimals):
        self.P = P
        self.molefrac = molefrac
        self.yeq = yeq
        self.xeq = xeq
        self.T_bubl = T_bubl
        self.T_dew = T_dew
        self.decimals = decimals

    def __repr__(self):
        d = self.decimals
        z = np.asarray(self.molefrac)

        if z.ndim == 1:  # single composition
            z_str = indent_array(z, tabs=0, precision=d)
            yeq = indent_array(self.yeq, tabs=0, precision=d)
            T_bubl = indent_array(self.T_bubl, tabs=0, precision=d)
            xeq = indent_array(self.xeq, tabs=0, precision=d)
            T_dew = indent_array(self.T_dew, tabs=0, precision=d)
            disp_text = ('Mixture VLE properties:\n'
                         f'    Pressure (P, bar): {self.P:.{d}f}\n'
                         f'    Overall composition (z): {z_str}\n'
                         f'    If this mixture were a saturated liquid:\n'
                         f'        Vapor in equilibrium (yeq): '
                         f'{yeq}\n'
                         f'        Bubble T (T_bubl, °C): '
                         f'{T_bubl}\n'
                         f'    If this mixture were a saturated vapor:\n'
                         f'        Liquid in equilibrium (xeq): '
                         f'{xeq}\n'
                         f'        Dew T (T_dew, °C): {T_dew} °C\n'
                         f'    Compositions are reported as mole fractions\n')
        else:  # multiple compositions
            z_str = indent_array(z, tabs=1, precision=d)
            yeq = indent_array(self.yeq, tabs=2, precision=d)
            T_bubl = indent_array(self.T_bubl, tabs=2, precision=d)
            xeq = indent_array(self.xeq, tabs=2, precision=d)
            T_dew = indent_array(self.T_dew, tabs=2, precision=d)
            disp_text = ('Mixture VLE properties:\n'
                         f'    Pressure (P, bar): {self.P:.{d}f}\n'
                         f'    Overall composition (z):\n{z_str}\n'
                         f'    If this mixture were a saturated liquid:\n'
                         f'        Vapor in equilibrium (yeq):\n{yeq}\n'
                         f'        Bubble T (T_bubl, °C):\n{T_bubl}\n'
                         f'    If this mixture were a saturated vapor:\n'
                         f'        Liquid in equilibrium (xeq):\n{xeq}\n'
                         f'        Dew T (T_dew, °C):\n{T_dew} °C\n'
                         f'    Compositions are reported as mole fractions\n')
        return disp_text


class StationaryPoints:
    """
    Container for the result of stationary point search.
    """

    def __init__(self, x, T, eigenvals, stability, decimals):
        self.x = x
        self.T = T
        self.eigenvals = eigenvals
        self.stability = stability
        self.decimals = decimals

    def __repr__(self):
        d = self.decimals
        disp_text = ''
        for i in range(len(self.T)):
            disp_row = (
                '[' +
                ', '.join(f'{self.x[i, j]:.{d}f}' for j in range(3)) + ']' +
                '  ' + f'{self.T[i]:{d + 4}.{d}f} °C' +
                '  eigenvalues: ' +
                '  '.join(f'{self.eigenvals[i, j]:+.{d}f}' for j in range(2)) +
                '  ' + (f'({self.stability[i]})\n')
            )
            disp_text = disp_text + disp_row
        return disp_text


class FlashResult:
    """
    Container for the result of a flash calculation.
    """

    def __init__(self, xF, x, y, T, P, Lfrac, Vfrac, decimals):
        self.xF = xF
        self.x = x
        self.y = y
        self.P = P
        self.T = T
        self.Lfrac = Lfrac
        self.Vfrac = Vfrac
        self.decimals = decimals

    def __repr__(self):
        d = self.decimals
        xF_str = indent_array(np.asarray(self.xF), tabs=0, precision=d)
        x_str = indent_array(np.asarray(self.x), tabs=0, precision=d)
        y_str = indent_array(np.asarray(self.y), tabs=0, precision=d)

        disp_text = ('Flash results:\n'
                     f'    Feed: {xF_str}\n'
                     f'    Liquid product (x): {x_str}\n'
                     f'    Vapor product (y): {y_str}\n'
                     f'    Temperature (T, °C): {self.T:.{d}f}\n'
                     f'    Pressure (P, bar): {self.P:.{d}f}\n'
                     f'    Liquid fraction (Lfrac): {self.Lfrac:.{d}f}\n'
                     f'    Vapor fraction (Vfrac): {self.Vfrac:.{d}f}\n'
                     f'    Compositions are reported as mole fractions\n')
        return disp_text


class DistillationResult:
    """
    Container for the result of a distillation calculation.
    """

    def __init__(self, xF, xD, xB, D, B, q, r, s,
                 nstg, fstg, xLr, yVr, Tr, xLs, yVs, Ts, P, decimals):
        self.xF = xF
        self.xD = xD
        self.xB = xB
        self.D = D
        self.B = B
        self.q = q
        self.r = r
        self.s = s
        self.nstg = nstg
        self.fstg = fstg
        self.xLr = xLr
        self.yVr = yVr
        self.Tr = Tr
        self.xLs = xLs
        self.yVs = yVs
        self.Ts = Ts
        self.P = P
        self.decimals = decimals

    def __repr__(self):
        d = self.decimals

        stage_text_r = "Stagewise results (R-section):\n"
        for i in range(len(self.Tr)):
            xLr_row = (
                '[' +
                '  '.join(f'{val:{d + 4}.{d}f}' for val in self.xLr[i])
                + ']'
            )
            yVr_row = (
                '[' +
                '  '.join(f'{val:{d + 4}.{d}f}' for val in self.yVr[i])
                + ']'
            )
            stage_text_r += f'    # {i+1:4d}  xL (liq): {xLr_row}  '
            stage_text_r += f"yV (vap): {yVr_row}  "
            stage_text_r += f"T: {self.Tr[i, 0]:{d + 4}.{d}f} °C\n"

        stage_text_s = "Stagewise results (S-section):\n"
        for i in range(len(self.Ts)):
            xLs_row = (
                '[' +
                '  '.join(f'{val:{d + 4}.{d}f}' for val in self.xLs[i])
                + ']'
            )
            yVs_row = (
                '[' +
                '  '.join(f'{val:{d + 4}.{d}f}' for val in self.yVs[i])
                + ']'
            )
            stage_text_s += f'    # {i+1:4d}  xL (liq): {xLs_row}  '
            stage_text_s += f"yV (vap): {yVs_row}  "
            stage_text_s += f"T: {self.Ts[i, 0]:{d + 4}.{d}f} °C\n"

        xF_str = indent_array(np.asarray(self.xF), tabs=0, precision=d)
        disp_text = ('Distillation parameters:\n'
                     f'    Feed: {xF_str}\n'
                     f'    Distillate composition (xD): {self.xD:.{d}f}\n'
                     f'    Bottoms composition (xB): {self.xB:.{d}f}\n'
                     f'    Distillate flow rate (D): {self.D:.{d}f}\n'
                     f'    Bottoms flow rate (B): {self.B:.{d}f}\n'
                     f'    Feed quality (q): {self.q:.{d}f}\n'
                     f'    Reflux ratio (r): {self.r:.{d}f}\n'
                     f'    Boilup ratio (s): {self.s:.{d}f}\n'
                     f'    Number of stages (nstg): {self.nstg}\n'
                     f'    Feed stage (fstg): {self.fstg}\n'
                     f'    Pressure (P, bar): {self.P:.{d}f}\n'
                     f'{stage_text_r}'
                     f'{stage_text_s}'
                     'Compositions are reported as mole fractions\n'
                     'Flows are molar relative to the feed\n')
        return disp_text


def normalize_fraction(x, tol=1e-4):
    """
    Normalize a ternary mole fraction array.

    Example:
        normalize_fraction([0.35, 0.20, 0.15], tol=1e-4)

    Args:
        x: An array of mole fractions, with components ordered
            according to the ``axes`` specification used when
            constructing the ``TernarySystem``. All three mole fractions
            must be specified and must add up to 1. This argument can
            also contain multiple composition points in the form of an
            n x 3 matrix or an m x n x 3 meshgrid, as long as the last
            dimension contains the three mole fractions.
        tol: Tolerance. Normalization is done if the sum of the mole
            fractions in x differs from 1 by more than the tolerance.

    Returns:
        An array of normalized mole fractions.
    """

    x = np.asarray(x)
    s = np.sum(x, axis=-1, keepdims=True)
    return np.where(np.abs(s - 1) <= tol, x, x / s)


def euclidean_distance(x, y):
    """
    Calculate Euclidean distance between two composition points x and y
        in composition space.

    Example:
        euclidean_distance([0.3, 0.40, 0.3], [0.2, 0.1, 0.7])

    Arguments:
        x, y: Arrays of mole fractions, with components ordered
            according to the ``axes`` specification used when
            constructing the ``TernarySystem``. All three mole fractions
            must be specified and must add up to 1. These arguments can
            also contain multiple composition points in the form of an
            n x 3 matrix or an m x n x 3 meshgrid, as long as the last
            dimension of each array contains the three mole fractions.

    Returns:
        A scalar denoting the Euclidean distance.
    """

    x = np.asarray(x)
    y = np.asarray(y)

    if x.shape[-1] != 3 or y.shape[-1] != 3:
        raise ValueError(
            'Input arrays must have exactly three mole fractions ' +
            'along their last dimension'
        )

    diff = x - y  # difference along last axis
    return np.sqrt(np.sum(diff * diff, axis=-1))


def out_of_triangle(t, y):
    """
    Determine if a 3-D composition point y is outside the mole fraction
        triangle defined by the points (0, 0, 1), (1, 0, 0) and (0, 1, 0).

    Note:
        This function is used in residue curve integration.
    """

    x1, x2, x3 = y
    # return < 0 if point is inside triangle, > 0 if outside
    return max(
        x1 - 1,
        0 - x1,
        x2 - 1,
        0 - x2,
        x3 - 1,
        0 - x3
    )


def out_of_triangle2(t, y):
    """
    Determine if a 2-D composition point y is outside the mole fraction
        triangle defined by the points (0, 0), (1, 0) and (0, 1).

    Note:
        This function is used in residue curve integration, especially
            when plotting separatrices.
    """

    x1, x2 = y
    # return < 0 if point is inside triangle, > 0 if outside
    return max(
        x1 - 1,
        0 - x1,
        x2 - 1,
        0 - x2
    )


def sign_tol(a, tol=1e-5):
    """
    Determines the sign of a scalar within a tolerance.
    """

    if a > tol:
        sign = 1
    elif a < -tol:
        sign = -1
    else:
        sign = 0
    return sign


def is_parallel(v1, v2, tol=1e-6):
    """
    Determine if two vectors are parallel.
    """

    # Test cross product component along perpendicular direction
    return abs(v1[0]*v2[1] - v1[1]*v2[0]) < tol


def indent_lines(text, tabs=1):
    """
    Add a specified number of tabs before each line of a multiline string.

    Note:
        This function is used in displaying VLE property, flash and
            distillation results.
    """

    prefix = "\t" * tabs
    return "\n".join(prefix + line for line in text.split("\n"))


def indent_array(array, tabs=1, **kwargs):
    """
    Convert an array to a string after applying any **kwargs, and pass to
        indent_lines, which displays the string with the specified number of
        tabs.

    Note:
        This function is used in displaying VLE property, flash and
            distillation results.
    """

    s = np.array2string(array, **kwargs)
    return indent_lines(s, tabs=tabs)



# Reaction Rate Field Plot: Reactors
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2017
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

import numpy as np
import scipy as sci
import warnings


class ReactorRegistry:
    def __init__(self):
        self._store = {}

    def add(self, reactor):
        name = reactor.name
        self._store[name] = reactor
        setattr(self, name, reactor)

    def __getitem__(self, key):
        return self._store[key]

    def __iter__(self):
        return iter(self._store)  # iterates keys

    def __len__(self):
        return len(self._store)

    def items(self):
        return self._store.items()

    def keys(self):
        return self._store.keys()

    def values(self):
        return self._store.values()

    def __repr__(self):
        if not self._store:
            return "Reactors()"
        lines = ["Reactors("]
        for name, obj in self._store.items():
            lines.append(f"{name}:\n{obj!r}")
        lines.append(")")
        return "\n".join(lines)


reactors = ReactorRegistry()


class Reactor:
    """
    Container for reactor interacting with a reactive system.

    Example:
        pfr1 = Reactor(flow_type='plug', feed=[1, 0], time=5, name='PFR1')

    Args:
        flow_type: Keyword indicating type of flow in the reactor.
            PFR options:  'pfr', 'plug', 'plug flow', 'tube'.
            CSTR options: 'cstr', 'mixed', 'tank', 'stirred tank',
            'continuously stirred tank'.
        feed: Feed coordinates, specified in the same sequence as axis
            components.
        time: Tau limit. The reactor is simulated from tau = 0 up to this time.
            Defaults to 10.
        name: Optional name for the reactor, e.g., 'PFR from fresh feed'.

    Returns:
        Reactor object containing the specified reactor, flow_type, feed and
            tau limit. After the reactor is simulated, the object also contains
            a curve object (Line2D or QuadContourSet), xy coordinates of the
            reactor curve and slopes of tangents along the reactor curve.
    """

    PFR_KEYWORDS = ('pfr', 'plug', 'plug flow', 'tube')
    CSTR_KEYWORDS = ('cstr', 'mixed', 'stirred tank', 'tank',
                     'continuously stirred tank')

    def __init__(self, flow_type, feed, time=10, name='',
                 curve=None, tau=None, x=None, y=None, xy=None,
                 slope_x=None, slope_y=None):
        self.flow_type = flow_type
        self.feed = feed
        self.time = time
        self.name = name
        self.curve = []
        self.tau = np.array([])
        self.x = np.array([])
        self.y = np.array([])
        self.xy = np.array([])
        self.slope_x = np.array([])
        self.slope_y = np.array([])
        self.segment = np.array([])
        self.simulated = False

        if self.flow_type not in self.PFR_KEYWORDS + self.CSTR_KEYWORDS:
            raise ValueError('flow_type does not match PFR or CSTR keywords')

        self.feed = np.asarray(self.feed)
        if not np.issubdtype(self.feed.dtype, np.number):
            feed_type_error = (
                'feed must be an array with integer or float elements'
            )
            raise TypeError(feed_type_error)
        if np.size(self.feed) < 2:
            feed_value_error = 'feed must be an array with two components'
            raise ValueError(feed_value_error)

        if self.time < 0:
            raise ValueError('Reactor time must be nonnegative')

        if reactors is not None:
            reactors.add(self)

    def __repr__(self):
        if np.array(self.tau).size == 0:
            disp_text = (f'Reactor:\n'
                         f'    name:\t{self.name}\n'
                         f'    flow type:\t{self.flow_type}\n'
                         f'    feed:\t{self.feed}')
        else:
            n_tau = np.array(self.tau).size
            disp_text = (f'Reactor:\n'
                         f'    name:\t{self.name}\n'
                         f'    flow type:\t{self.flow_type}\n'
                         f'    feed:\t{self.feed}\n'
                         f'    time:\t{self.time}\n'
                         f'    curve:\tarray of curve objects\n'
                         f'    tau:\t{n_tau}×1 numpy array of tau values\n'
                         f'    x:\t\t{n_tau}×1 numpy array of x values\n'
                         f'    y:\t\t{n_tau}×1 numpy array of y values\n'
                         f'    xy:\t\t{n_tau}×2 numpy array of [x, y] values\n'
                         f'    slope_i:\t{n_tau}×1 numpy arrays of slope data')
        return disp_text

    def simulate(self, system, ax, time_limit=0,
                 n_points=1000, show_feed=False,
                 inequality_tol=1e-6, **kwargs):
        """
        Simulate and plot a reactor.

        Example:
            pfr1.simulate(sysABC, ax, time_limit=20, n_points=1000,
                          show_feed=True)

        Args:
            system: Reaction system operating in the reactor.
            ax: Axes on which the reactor curve is to be plotted.
            time_limit: Tau limit. The reactor is simulated from tau = 0 up to
                this time. Overrides the ``time`` argument of the reactor.
            n_points: Number of integration or calculation points. For PFRs,
                this value equals the exact number of points in the curve.
                For CSTRs, it is the number of points along each dimension for
                which the contour is evaluated. Thus, the exact number of
                points in the CSTR curve will be proportional to, but will
                differ from n_points.
            show_feed: True/False value indicating whether feed should be
                displayed as a point.
            inequality_tol: Tolerance for any inequalities, defaults to 1e-6.
            **kwargs: Additional keyword arguments passed to matplotlib.pyplot.
                Common options include ``linewidth`` for the reactor curve and
                ``marker`` for the feed point(s). Note that PFR curves and
                feeds are always plotted in red (#ff0000), whereas CSTR curves
                and feeds are always plotted in blue (#0000ff).

        Returns:
            Axes passed to the function and an updated reactor object
                containing a curve object (Line2D or QuadContourSet), xy
                coordinates of the reactor curve and slopes of tangents along
                the reactor curve.
        """

        if time_limit < 0:
            raise ValueError('time_limit must be nonnegative')

        if n_points < 0:
            raise ValueError('n_points must be nonnegative')

        if not isinstance(n_points, int):
            raise TypeError('n_points must be an integer')

        if not isinstance(show_feed, bool):
            raise TypeError('show_feed must be Boolean (True or False)')

        # Parse kwargs
        if self.flow_type.lower() in self.PFR_KEYWORDS:
            forced_line_kwargs = {'linestyle': '-', 'color': 'r'}
        elif self.flow_type.lower() in self.CSTR_KEYWORDS:
            forced_line_kwargs = {'linestyle': '-', 'color': 'b'}

        conflict_keys = {'linestyle', 'ls',
                         'color', 'c'}  # may conflict with user's keys
        for k in conflict_keys:  # remove conflicted keys from kwargs
            kwargs.pop(k, None)

        common_keys = {'alpha'}  # shared between point and line

        point_keys = {
                'marker',
                'markersize', 'ms',
                'markerfacecolor', 'mfc',
                'markeredgecolor', 'mec',
                'markeredgewidth', 'mew',
                'fillstyle',
            }

        line_keys = {
            'linewidth', 'lw',
            'drawstyle',
            'dash_capstyle',
            'dash_joinstyle',
            'solid_capstyle',
            'solid_joinstyle',
            'antialiased',
        }

        common_kwargs, point_kwargs, line_kwargs = {}, {}, {}

        for key, value in kwargs.items():
            if key in common_keys:
                common_kwargs[key] = value
            elif key in point_keys:
                point_kwargs[key] = value
            elif key in line_keys:
                line_kwargs[key] = value
            else:
                pass

        point_kwargs = {**common_kwargs, **point_kwargs}
        line_kwargs = {**common_kwargs, **forced_line_kwargs, **line_kwargs}

        def ratet(t, x):
            return system.rate(x)

        xF = self.feed
        if np.size(self.feed, 0) != 2:
            raise ValueError('The feed should have exactly 2 coordinates')
        if show_feed:
            ax.plot(xF[0], xF[1], **point_kwargs)

        if self.flow_type.lower() in self.PFR_KEYWORDS:
            if time_limit == 0:
                int_t = self.time
                time_limit = self.time
            else:
                int_t = time_limit

            with np.errstate(all='ignore'):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', RuntimeWarning)
                    res = sci.integrate.solve_ivp(ratet, [0, int_t], xF,
                                                  method='DOP853',
                                                  dense_output=True)
                    t_vals = np.linspace(0, int_t, n_points)
                    x_vals = res.sol(t_vals)[0]
                    x_uniform = np.linspace(x_vals.min(),
                                            x_vals.max(),
                                            n_points)
                    self.x = x_uniform

                    t_from_x = sci.interpolate.interp1d(
                        x_vals, t_vals, fill_value="extrapolate"
                    )
                    t_uniform = t_from_x(x_uniform)
                    self.tau = t_uniform

                    y_uniform = res.sol(t_uniform)
                    self.y = y_uniform[1, :]
                    self.xy = np.column_stack((self.x, self.y))
                    self.curve = ax.plot(self.x, self.y, **line_kwargs)

                    rates_pfr = np.array([system.rate(p) for p in self.xy])
                    self.slope_x = rates_pfr[:, 0]
                    self.slope_y = rates_pfr[:, 1]
                    self.segment = np.zeros_like(self.x)  # only segment 0
                    self.simulated = True

        if self.flow_type.lower() in self.CSTR_KEYWORDS:
            if time_limit == 0:
                calc_t = self.time
                time_limit = self.time
            else:
                calc_t = time_limit

            h_lim = system.h_lim
            v_lim = system.v_lim
            xx = np.linspace(h_lim[0], h_lim[1], n_points)
            yy = np.linspace(v_lim[0], v_lim[1], n_points)
            X, Y = np.meshgrid(xx, yy)

            def cstrpt(x, y):
                r = system.rate([x, y])
                with np.errstate(all='ignore'):
                    f = y - xF[1] - r[1] / r[0] * (x - xF[0])
                    return f

            self.curve = ax.contour(X, Y, cstrpt(X, Y),
                                    levels=[0], linewidths=0)

            xs, ys, xys = [], [], []
            dxs, dys, segment = [], [], []
            seg_count = 0  # global segment counter

            for level in self.curve.allsegs:
                for seg in level:
                    if seg.size == 0:
                        continue
                    seg_count += 1  # increment counter

                    xy = seg
                    xseg = seg[:, 0]
                    yseg = seg[:, 1]
                    dx = np.gradient(xseg)
                    dy = np.gradient(yseg)

                    xys.append(xy)
                    xs.append(xseg)
                    ys.append(yseg)
                    dxs.append(dx)
                    dys.append(dy)
                    segment.append(seg_count * np.ones_like(xseg))

                    # Add NaN breaks between segments
                    xys.append(np.array([[np.nan, np.nan]]))
                    xs.append(np.array([np.nan]))
                    ys.append(np.array([np.nan]))
                    dxs.append(np.array([np.nan]))
                    dys.append(np.array([np.nan]))
                    segment.append(np.array([seg_count + 0.5]))  # in between

            xy = np.transpose(np.vstack(xys)) if xys else np.empty((2, 0))
            self.x = np.concatenate(xs) if xs else np.array([])
            self.y = np.concatenate(ys) if ys else np.array([])
            self.slope_x = np.concatenate(dxs) if dxs else np.array([])
            self.slope_y = np.concatenate(dys) if dys else np.array([])
            self.segment = np.concatenate(segment) if segment else np.array([])

            with np.errstate(invalid='ignore', divide='ignore'):
                self.tau = np.array(
                    (xy - np.array(xF)[:, None]) /
                    system.rate([xy[0, :], xy[1, :]])
                )[0, :]
                # tau is calculated in duplicate from both axes, so discard one

            # Eliminate points at which tau < 0 or tau > specified time limit
            mask = (self.tau < 0) | (self.tau > calc_t)
            self.x[mask] = np.nan
            self.y[mask] = np.nan
            self.slope_x[mask] = np.nan
            self.slope_y[mask] = np.nan
            self.tau[mask] = np.nan

            # Eliminate points at which CSTR design equation is not satisfied
            # This step should remove any contour artifacts
            tol = 1e-2
            r = system.rate([self.x, self.y])
            dy = np.abs(self.y - xF[1] - self.tau * r[1])
            dx = np.abs(self.x - xF[0] - self.tau * r[0])
            self.x = np.where(dy + dx > tol, np.nan, self.x)
            self.y = np.where(dy + dx > tol, np.nan, self.y)
            self.tau = np.where(dy + dx > tol, np.nan, self.tau)
            self.slope_x = np.where(dx > tol, np.nan, self.slope_x)
            self.slope_y = np.where(dx > tol, np.nan, self.slope_y)

            # Eliminate all nans except those between segments
            # In between segments, self.segment will be 0.5, 1.5, etc.
            mask = ~(
                (np.isnan(self.x) |
                 np.isnan(self.y) |
                 np.isnan(self.tau)) &
                (np.mod(self.segment, 1) == 0)
            )
            self.x = self.x[mask]
            self.y = self.y[mask]
            self.tau = self.tau[mask]
            self.slope_x = self.slope_x[mask]
            self.slope_y = self.slope_y[mask]
            self.segment = self.segment[mask]

            self.xy = np.column_stack((self.x, self.y))
            self.curve = ax.plot(self.x, self.y, **line_kwargs)

            self.simulated = True
        return ax

    def plot_point(self, ax, x=[], y=[], tau=[],
                   annotate=True, annotation='all', **kwargs):
        """
        Locate and plot a point corresponding to a specified value of x, y
            and/or tau on a reactor curve.

        Notes:
            More than one of x, y or tau could be specified.
            If an exact match is not found, the nearest point on the curve
                will be found.
            It may be useful to duplicate the figure first, especially when
                trying out points that are not yet final:

                fig1, ax1 = duplicate_figure(fig)

        Examples:
            pfr1.plot_point(ax1, x=0.6)
            ax1, point, text = pfr1.plot_point(ax1, x=0.6,
                                               ha='right',
                                               va='bottom')
            pfr1.plot_point(ax1, y=0.2, markersize=6, annotate=False)
            cstr1.plot_point(ax1, tau=0.5, ha='left', va='top')
            fig1  # display figure after points are plotted

        Args:
            ax: Axes on which the point should be plotted.
            x, y, tau: Scalars specifying the point. At least one
                must be provided. See notes above.
            annotate (verb): Boolean parameter indicating whether the point
                should be annotated with values of x, y and/or tau.
            annotation (noun): String indicating the type of annotation for
                the point. Options are 'tau', 'x', 'y' 'xy', 'all'. Default
                is 'all'. Not  case-sensitive.
            **kwargs: Additional keyword arguments passed to
                matplotlib.pyplot.plot and  matplotlib.pyplot.text.
                Common options include ``marker`` and ``markersize`` for
                the point and ``fontsize`` (``size``), ``fontstyle``
                (``style``), ``fontweight`` (``weight``),
                ``horizontalalignment`` (``ha``) and
                ``verticalalignment`` (``va``) for the text. Points on PFRs
                are always plotted in red (#ff0000), whereas points on CSTRs
                are always plotted in blue (#0000ff). Points are always
                annotated in black (#000000). Users may change the colors
                and markers by calling:
                    ax1, point, text = reactor.plot_point(...)
                and editing the attributes:
                    point.marker = '+'
                    text.color = 'b'

        Returns:
            Axes: Axes passed to the function.
            Point: Line2D object containing the plotted point.
            Text: If annotate is set to True, text object containing the
                annotation.
        """
        if not self.simulated:
            print('Reactor must be simulated before plotting point(s)')
            return
        else:
            if (
                not np.isscalar(x) and
                not np.isscalar(y) and
                not np.isscalar(tau)
            ):
                raise ValueError(
                    'At least one of x, y or tau should be provided'
                )

            # Parse kwargs
            if self.flow_type in self.PFR_KEYWORDS:
                forced_kwargs = {'color': 'r'}
            elif self.flow_type in self.CSTR_KEYWORDS:
                forced_kwargs = {'color': 'b'}
            forced_point_kwargs = {'marker': 'o'}
            forced_text_kwargs = {'color': 'k'}

            for k in ('color', 'c'):  # conflicted kwarg
                kwargs.pop(k, None)

            common_keys = {'alpha'}

            point_keys = {
                'marker',
                'markersize', 'ms',
                'markerfacecolor', 'mfc',
                'markeredgecolor', 'mec',
                'markeredgewidth', 'mew',
                'fillstyle',
            }

            text_keys = {
                'fontsize', 'fontweight', 'fontstyle',
                'ha', 'va', 'rotation',
                'bbox',
                'family', 'name',
            }

            common_kwargs, point_kwargs, text_kwargs = {}, {}, {}

            for key, value in kwargs.items():
                if key in common_keys:
                    common_kwargs[key] = value
                elif key in point_keys:
                    point_kwargs[key] = value
                elif key in text_keys:
                    text_kwargs[key] = value
                else:
                    pass

            point_kwargs = {**forced_kwargs,
                            **forced_point_kwargs,
                            **common_kwargs,
                            **point_kwargs}
            text_kwargs = {**forced_kwargs,
                           **forced_text_kwargs,
                           **common_kwargs,
                           **text_kwargs}

            mask = (
                ~np.isnan(self.x)
                & ~np.isnan(self.y)
                & ~np.isnan(self.tau))

            xx = self.x[mask]
            yy = self.y[mask]
            tt = self.tau[mask]

            ssr = np.zeros(np.size(xx))
            if np.isscalar(x):
                ssr += (xx - x)**2
            if np.isscalar(y):
                ssr += (yy - y)**2
            if np.isscalar(tau):
                ssr += (tt - tau)**2
            idx = ssr.argmin()

            x = xx[idx]
            y = yy[idx]
            tau = tt[idx]

            label_tau = f' $\\tau=$ {np.round(tau, 3)}'
            label_xy = f'({np.round(x, 3)}, {np.round(y, 3)})'
            label_x = f'x = {np.round(x, 3)}'
            label_y = f'y = {np.round(y, 3)}'
            label_all = label_tau + ', ' + label_xy

            if annotation.lower() == 'tau':
                label = label_tau
            elif annotation.lower() == 'xy':
                label = label_xy
            elif annotation.lower() == 'x':
                label = label_x
            elif annotation.lower() == 'y':
                label = label_y
            elif annotation.lower() == 'all':
                label = label_all
            else:
                raise ValueError(
                    f"Point annotation should be 'tau', 'xy, 'x', 'y'"
                    f"or 'all' (case-insensitive), got '{annotation}'"
                )

            if self.flow_type in self.PFR_KEYWORDS:
                point = ax.plot(x, y, **point_kwargs)
                if annotate:
                    text = ax.text(x, y, label, **text_kwargs)
                    return ax, point, text
                else:
                    return ax, point
            elif self.flow_type in self.CSTR_KEYWORDS:
                point = ax.plot(x, y, **point_kwargs)
                if annotate:
                    text = ax.text(x, y, label, **text_kwargs)
                    return ax, point, text
                else:
                    return ax, point

    def plot_tangent(self, ax, from_point, near_point=[],
                     distance_x=1, distance_h=None,
                     distance_y=1, distance_v=None,
                     tol=1e-2, **kwargs):
        """
        Plot a tangent to a reactor curve from a specified point.

        Notes:
            The tangent could optionally be constrained to touch the curve
                within a specified horizontal or vertical distance from
                another point.
            It may be useful to duplicate the figure first, especially when
                trying out tangents that are not yet final:

                fig1, ax1 = duplicate_figure(fig)

        Example:
            reactor.plot_tangent(ax,
                                 from_point=[1, 0.2],
                                 near_point=[0.4, 0.3],
                                 distance_v=0.05)

        Args:
            ax: Axes on which the tangent should be plotted
            from_point: Point from where tangent should be plotted.
            near_point: Point to which tangent's point of contact should be
                near. Defaults to reactor feed.
            distance_h or distance_x: Horizontal distance from near_point
                to the point of contact. Defaults to 1.
            distance_v or distance_y: Vertical distance from near_point
                to the point of contact. Defaults to 1.
            tol: Tolerance for tangent. Defaults to 1e-2.
            **kwargs: Additional keyword arguments passed to
                matplotlib.pyplot.plot. Common options include ``color``
                (``c``), ``linewidth`` (``lw``) and ``linestyle`` (``ls``).

        Returns:
            Axes: Axes passed to the function.
            Line: If a tangent was found, Line2D object containing the tangent.
        """

        if not self.simulated:
            print('Reactor must be simulated before plotting tangent')
            return
        else:
            if distance_x is None and distance_h is not None:
                distance_x = distance_h
            elif distance_h is not None and distance_x is not None:
                raise ValueError(
                    'Use either distance_h or distance_x, not both'
                )
            if distance_x is None:
                distance_x = 1

            if distance_y is None and distance_v is not None:
                distance_y = distance_v
            elif distance_v is not None and distance_y is not None:
                raise ValueError(
                    "Use either distance_v or distance_y, not both"
                )
            if distance_y is None:
                distance_y = 1

            from_point = np.array(from_point)
            if near_point:
                near_point = np.array(near_point)
            else:
                near_point = self.feed  # defaults to reactor feed

            #  Filter for distance_x, then for distance_y
            d_x = np.abs(self.x - near_point[0])
            xx = np.where(d_x < distance_x, self.x, np.nan)
            yy = np.where(d_x < distance_x, self.y, np.nan)
            sl_x = np.where(d_x < distance_x, self.slope_x, np.nan)
            sl_y = np.where(d_x < distance_x, self.slope_y, np.nan)

            d_y = np.abs(self.y - near_point[1])
            xx = np.where(d_y < distance_y, xx, np.nan)
            yy = np.where(d_y < distance_y, yy, np.nan)
            sl_x = np.where(d_y < distance_y, sl_x, np.nan)
            sl_y = np.where(d_y < distance_y, sl_y, np.nan)

            angle_point_to_reactor = np.atan2(yy - from_point[1],
                                              xx - from_point[0])
            angle_tangent = np.atan2(sl_y, sl_x)
            delta = angle_point_to_reactor - angle_tangent

            if delta.any():
                sin_delta = np.abs(np.sin(delta))
                xx = xx[~np.isnan(sin_delta)]
                yy = yy[~np.isnan(sin_delta)]
                sin_delta = sin_delta[~np.isnan(sin_delta)]

                idx = sin_delta.argmin()
                if sin_delta[idx] < tol:
                    (line,) = ax.plot((from_point[0], xx[idx]),
                                      (from_point[1], yy[idx]), **kwargs)
                    return ax, line
                else:
                    print(
                        f'Tangent to {self.name} not found from '
                        f'{from_point} within tolerance'
                        )
                    return ax, None
            else:
                print(f'Tangent to {self.name} not found from '
                      f'{from_point} within tolerance'
                      )
                return ax, None


def common_tangent(ax, reactor1, reactor2,
                   near_point=[], distance_x=None, distance_h=None,
                   distance_y=None, distance_v=None, tol=1e-2,
                   **kwargs):
    """
    Plots a common tangent to a pair of reactors, if one exists.

    Example:
        common_tangent(ax, reactor1, reactor2, near_point=[0.4, 0.3],
        distance_h=0.3, tol=1e-2, **kwargs)

    Args:
        ax: Axes on which the tangent is to be plotted.
        reactor1, reactor2: Reactors, which should already be simulated.
        near_point: Point to which tangent's point of contact should be
            near. Defaults to reactor feed.
        distance_h or distance_x: Horizontal distance from near_point
            to the point of contact. Defaults to 1.
        distance_v or distance_y: Vertical distance from near_point
            to the point of contact. Defaults to 1.
        tol: Tolerance for tangent. Defaults to 1e-2.
        **kwargs: Additional keyword arguments passed to
            matplotlib.pyplot.plot. Common options include ``color``
            (``c``), ``linewidth`` (``lw``) and ``linestyle`` (``ls``).

    Returns:
        Axes: Axes passed to the function.
        Line: If a common tangent was found, Line2D object containing the
            tangent.
    """
    if (not reactor1.simulated) or (not reactor2.simulated):
        print('Both reactors must be simulated before plotting common tangent')
        return
    else:
        if distance_x is None and distance_h is not None:
            distance_x = distance_h
        elif distance_h is not None and distance_x is not None:
            raise ValueError("Use either distance_h or distance_x, not both")
        if distance_x is None:
            distance_x = 1

        if distance_y is None and distance_v is not None:
            distance_y = distance_v
        elif distance_v is not None and distance_y is not None:
            raise ValueError("Use either distance_v or distance_y, not both")
        if distance_y is None:
            distance_y = 1

        if near_point:
            near_point = np.array(near_point)
        else:
            near_point = reactor1.feed  # defaults to first reactor's feed

        # Remove feed points
        x1 = reactor1.x[1:]
        y1 = reactor1.y[1:]
        m1 = reactor1.slope_y[1:] / reactor1.slope_x[1:]
        x2 = reactor2.x[1:]
        y2 = reactor2.y[1:]
        m2 = reactor2.slope_y[1:] / reactor2.slope_x[1:]

        #  Filter for distance_x, then for distance_y
        d_x = np.abs(x1 - near_point[0])
        x1 = np.where(d_x < distance_x, x1, np.nan)
        y1 = np.where(d_x < distance_x, y1, np.nan)
        m1 = np.where(d_x < distance_x, m1, np.nan)
        d_y = np.abs(y1 - near_point[1])
        x1 = np.where(d_y < distance_y, x1, np.nan)
        y1 = np.where(d_y < distance_y, y1, np.nan)
        m1 = np.where(d_y < distance_y, m1, np.nan)

        d_x = np.abs(x2 - near_point[0])
        x2 = np.where(d_x < distance_x, x2, np.nan)
        y2 = np.where(d_x < distance_x, y2, np.nan)
        m2 = np.where(d_x < distance_x, m2, np.nan)
        d_y = np.abs(y2 - near_point[1])
        x2 = np.where(d_y < distance_y, x2, np.nan)
        y2 = np.where(d_y < distance_y, y2, np.nan)
        m2 = np.where(d_y < distance_y, m2, np.nan)

        x1 = x1[~np.isnan(x1)]
        y1 = y1[~np.isnan(y1)]
        m1 = m1[~np.isnan(m1)]
        x2 = x2[~np.isnan(x2)]
        y2 = y2[~np.isnan(y2)]
        m2 = m2[~np.isnan(m2)]

        X1 = x1[:, None]
        Y1 = y1[:, None]
        M1 = m1[:, None]
        X2 = x2[None, :]
        Y2 = y2[None, :]
        M2 = m2[None, :]

        dx = X2 - X1
        dy = Y2 - Y1

        # Avoid division by zero
        invalid = np.isclose(dx, 0.0)
        slope_line = np.empty_like(dx)
        slope_line[~invalid] = dy[~invalid] / dx[~invalid]
        slope_line[invalid] = np.nan

        # Mismatch between line slope and each curve's local slope
        err1 = M1 - slope_line  # curve 1 slope vs line slope
        err2 = M2 - slope_line  # curve 2 slope vs line slope

        # Total squared error
        err = err1**2 + err2**2
        err[invalid] = np.inf

        # Best indices
        flat_idx = np.argmin(err)
        i_best, j_best = np.unravel_index(flat_idx, err.shape)

        if tol < np.inf and err[i_best, j_best] > tol:
            print("No common tangent found within tolerance")
            return
        else:
            line = ax.plot([x1[i_best], x2[j_best]],
                           [y1[i_best], y2[j_best]], **kwargs)
            return ax, line


def convexify(ax, boundaries, **kwargs):
    """ Plots the convex hull of a given set of curves (boundaries), which
    could be reactors or line objects.

    Example:
        convexify(ax, boundaries, alpha=0.2, **kwargs)

    Args:
        ax: axes of the figure on which the AR is to be plotted
        boundaries: list containing reactor objects and line objects

    Returns:
        Axes: Axes passed to the function.
        Hull: ConvexHull containing mathematical description of the hull.
    """

    reactors_simulated = True
    for boundary in boundaries:
        if isinstance(boundary, Reactor):
            if not boundary.simulated:
                reactors_simulated = False
    if reactors_simulated is False:
        print('All reactors must be simulated before convexifying')
        return
    else:
        # Parse kwargs
        common_keys = {'color', 'c', 'alpha'}

        line_keys = {
            'linestyle', 'ls',
            'linewidth', 'lw',
            'drawstyle',
            'dash_capstyle',
            'dash_joinstyle',
            'solid_capstyle',
            'solid_joinstyle',
            'antialiased',
            'marker',
            'markersize', 'ms',
            'markerfacecolor', 'mfc',
            'markeredgecolor', 'mec',
            'markeredgewidth', 'mew',
            'fillstyle',
        }

        fill_keys = {
            'edgecolor',
            'facecolor',
            'linewidth',
            'linestyle',
            'hatch',
            'antialiased',
            'rasterized',
            'path_effects',
        }

        common_kwargs, line_kwargs, fill_kwargs = {}, {}, {}

        for key, value in kwargs.items():
            if key in common_keys:
                common_kwargs[key] = value
            elif key in line_keys:
                line_kwargs[key] = value
            elif key in fill_keys:
                fill_kwargs[key] = value
            else:
                pass

        line_kwargs = {**common_kwargs, **line_kwargs}
        fill_kwargs = {**common_kwargs, **fill_kwargs}

        pts_list = []
        for boundary in boundaries:
            if isinstance(boundary, Reactor):
                xy = boundary.xy
            else:
                xy = np.column_stack((boundary.get_xdata(),
                                      boundary.get_ydata()))
            pts_list.append(xy)
        pts = np.vstack(pts_list)
        pts = pts[~np.isnan(pts).any(axis=1)]  # remove NaNs and duplicates
        pts = np.unique(pts, axis=0)

        hull = sci.spatial.ConvexHull(pts)
        hull_xy = pts[hull.vertices]
        ax.plot(hull_xy[:, 0], hull_xy[:, 1], **line_kwargs)  # hull boundary
        ax.fill(hull_xy[:, 0], hull_xy[:, 1], **fill_kwargs)  # hull
        return ax, hull



# Linear Programming: Computation of Simplex Tableaux
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu


def simplex_tableaux(c, A_ub=[], b_ub=[],
                     var_names=[], slack_names=[],
                     tol=1e-6, decimals=2, dec=2, iters=10,
                     colormap='Greys'):
    """
    Solve an LP by the simplex method, implemented as a series of matrix
        operations.

    Notes:
        This version:
            - Supports inequality constraints only; equality constraints
                are not accepted.
            - Does not update the auxiliary after each pivot.

    Example:
        Tlist = simplex_tableaux(c=[0,-1/2],
                                 A_ub=[[1,-1/2],[-1/3,1/2]],
                                 b_ub=[0,1],
                                 var_names=['x1', 'x3'],
                                 slack_names=['x4','x5'],
                                 tol=1e-6,
                                 decimals=3,
                                 colormap='YlGn')

    Args:
        c: Cost vector specifying objective function.
        A_ub: Matrix containing the coefficients of x_i in the inequality
            constraints (A_ub.x <= b_ub). The number of columns of A_ub
            should equal the number of elements in c. The number of rows
            of A_ub should equal the number of elements in b_ub.
        b_ub: Vector containing the constants in the inequality constraints
            (A_ub.x <= b_ub).
        var_names: Optional column headers in simplex tableaux, specified as
            a list of of strings equal to the number of variables x_i or the
            number of columns in A_ub. If this arg is not specified, these
            headers are created during execution.
        slack_names: Optional row headers in simplex tableaux, specified as
            a list of strings equal to the number of slack variables or the
            number of rows in A_ub. If this arg is not specified, these
            headers are created during execution.
        tol: Tolerance for element of ``z`` row that determines whether
            iterations will continue. Use this arg to suppress iterations
            with little or no incremental benefit. Defaults to 1e-6.
        iters: Number of tableaux to be computed beyond the initial tableau.
            Use this arg to suppress iterations with little or no incremental
            benefit. Defaults to 10.
        decimals: Number of decimals in output. Can also be specified as
            ``dec``. If both ``decimals`` and ``dec`` are specified,
            ``decimals`` overrides. Defaults to 2.
        colormap: Colormap used for conditional formatting of tableaux.
            Choose from the palettes available in Matplotlib:
            https://matplotlib.org/stable/users/explain/colors/colormaps.html.
            Defaults to 'Greys'.

    Returns:
        A list of pandas dataframes, each containing a simplex tableau with
            headers.
    """

    import numpy as np
    import pandas as pd

    # Examine vectors and matrix dimensions for consistency
    Arows = np.size(A_ub, axis=0)
    Acols = np.size(A_ub, axis=1)
    brows = np.size(b_ub)
    ccols = np.size(c)

    if Acols != ccols:
        raise ValueError(
            'Check input dimensions. ' +
            'A and c have different numbers of columns.'
        )
    if Arows != brows:
        raise ValueError(
            'Check input dimensions. ' +
            'A and b have different numbers of rows.'
        )

    if decimals is None:
        decimals = dec

    # Create headers (variable and slack names) if not provided
    if var_names == []:  # create var_names if empty
        var_names = [f"x{j}" for j in range(1, 1 + Acols)]
        # Acols and ccols would have been checked for consistency by now

    var_num = np.size(var_names)
    var_max = max(int(s.lstrip('x')) for s in var_names)

    if slack_names == []:  # create slack_names if empty
        slack_names = [f"x{i}" for i in range(var_max + 1,
                                              var_max + 1 + Arows)]
        # Arows and brows would have been checked for consistency by now

    slack_min = min(int(s.lstrip('x')) for s in slack_names)

    # Create slack_names if empty or index is <= var_max
    if slack_names == [] or slack_min <= var_max:
        slack_names = [f"x{i}" for i in range(var_max + 1,
                                              var_max + 1 + Arows)]
        # Arows and brows would have been checked for consistency by now

    slack_num = np.size(slack_names)

    if Acols != var_num:
        raise ValueError(
            'Check input dimensions. ' +
            'The number of original variables is different ' +
            'from the number of columns of A.'
        )
    if Arows != slack_num:
        raise ValueError(
            'Check input dimensions. ' +
            'The number of slack variables is different ' +
            'from the number of rows of A.'
        )

    # Create initial tableau
    n = np.size(c)  # number of original variables
    m = np.size(b_ub)  # number of slack variables

    # Function to get the ith row of an m x m identity matrix
    def sleq(i, m):
        identity = np.identity(m)
        return identity[i]

    # Construct initial tableau T
    T = np.concatenate(([1], c, np.zeros(m), [0]), 0)

    for i in range(0, m):
        t = np.concatenate(([0],
                            A_ub[i],
                            sleq(i, m),
                            [b_ub[i]]),
                           0)  # current row

        T = np.vstack((T, t))  # append current row below T

    Trows = np.size(T, axis=0)
    Tcols = np.size(T, axis=1)

    aux_needed = False  # compose auxiliary if needed

    c_aux = np.zeros(Tcols)  # initialize cost vector for auxiliary
    for i in range(1, 1 + m):
        if T[i, Tcols - 1] < 0:
            aux_needed = True
            c_aux += T[i]  # update cost vector for auxiliary objective

    if aux_needed is True:
        T = np.insert(T, 1, c_aux, axis=0)
        t = np.concatenate(([0], [1], np.zeros(m)), 0)
        T = np.insert(T, 1, t, axis=1)

    if aux_needed is False:
        row_names = ["z"]
    else:
        row_names = ["z", "z'"]

    for k in range(0, np.size(slack_names)):
        row_names.append(slack_names[k])

    if aux_needed is False:
        col_names = ["z"]
    else:
        col_names = ["z", "z'"]

    for j in range(0, np.size(var_names)):
        col_names.append(var_names[j])
    for i in range(0, np.size(slack_names)):
        col_names.append(slack_names[i])
    col_names.append('b')

    Tlist = [pd.DataFrame(T, columns=col_names,
                          index=row_names)]  # store tableau in a list

    # Use pandas to display T as a tableau
    def conditionally_format_initial(styler):
        styler.background_gradient(axis=None,
                                   vmin=np.min(T),
                                   vmax=np.max(T),
                                   cmap=colormap)
        return styler

    df = pd.DataFrame(T, columns=col_names,
                      index=row_names)
    df_style = df.style.format(
        precision=decimals).pipe(conditionally_format_initial).set_properties(
        **{'text-align': 'center'})

    df_style.set_table_styles([{'selector': 'th.col_heading',
                                'props': 'text-align: center;'},],
                              overwrite=False)
    show_df(df_style)
    N = 0  # iteration number

    Trows = np.size(T, axis=0)
    Tcols = np.size(T, axis=1)

    if aux_needed is True:
        start = 2
        index = 1
    else:
        start = 1
        index = 0

    # Iterate
    stop_condition = False
    while stop_condition is False:
        if aux_needed is True and T[1, Tcols - 1] >= 0:
            aux_needed = False
            index = 0

        pcol = 0  # initialize pivot column
        cval = tol

        for j in range(start, start + m + n):
            if T[index, j] < tol and T[index, j] < cval:
                cval = T[index, j]
                pcol = j  # identify pivot column
        if np.abs(cval) <= tol or N >= iters:
            stop_condition = True

        if stop_condition is False:
            prow = 0  # initialize pivot row
            rval = np.inf
            for i in range(start, start + m):
                if T[i, pcol] > 0 and T[i, Tcols - 1]/T[i, pcol] < rval:
                    rval = T[i, Tcols - 1] / T[i, pcol]
                    prow = i  # identify pivot row

            # Perform pivot
            # Initialize pivoting matrix R to identity matrix
            R = np.identity(Trows)

            # Compile factors to turn pivot column elements above and
            # below pivot element to zero
            factors = -T[:, pcol] / T[prow, pcol]
            factors[prow] = 1  # pivot row is unaltered during pivot
            R[:, prow] = factors

            T = np.dot(R, T)  # pivot operation

            row_names[prow] = col_names[pcol]  # update row headers

            # Append tableau to the list
            Tlist.append(pd.DataFrame(T, columns=col_names,
                                      index=row_names))

            def conditionally_format_current(styler):
                styler.background_gradient(axis=None,
                                           vmin=np.min(T),
                                           vmax=np.max(T),
                                           cmap=colormap)
                return styler

            df = pd.DataFrame(T, columns=col_names,
                              index=row_names)
            df_style = (
                df.style
                .format(precision=decimals)
                .pipe(conditionally_format_current)
                .set_properties(**{'text-align': 'center'})
            )
            df_style.set_table_styles([{'selector': 'th.col_heading',
                                        'props': 'text-align: center;'},],
                                      overwrite=False)
            show_df(df_style)
            N += 1

    return Tlist  # return tableaux as a list of dataframes


def show_df(df):
    """
    Display a DataFrame nicely in Jupyter, or print it in text mode.
    """
    try:
        from IPython.display import display
        display(df)
    except Exception:
        print(df.to_string())



# Heat Integration: Heat Exchanger Network Synthesis
# Translated from GAMS for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2017
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

# This model designs a heat exchanger network operating at minimal
# annual cost and satisfying specified heating and cooling requirements.
# The superstructure consists of k stages and up to 10 possible exchangers.
# The original model was written based on this article:
#     Yee, T F, and Grossmann, I E, Simultaneous Optimization of Models for
#     Heat Integration: Heat Exchanger Network Synthesis. Computers and
#     Chemical Engineering 14, 10 (1990), 1151-1184.
# The GAMS version of this model is available in the GAMS model library:
#     https://www.gams.com/47/gamslib_ml/libhtml/gamslib_synheat.html

# Keywords
#    heat exchanger network
#    mixed integer nonlinear programming
#    optimization
#    chemical engineering design

import pandas as pd
import gamspy as gams


def synheat(streams, DT_min=10, n_stages=3, stages=None,
            T_hot_util_in=450, T_hot_util_out=450,
            T_cold_util_in=20, T_cold_util_out=20,
            hot_util_cost=100, cold_util_cost=100,
            exchanger_fixed_charge=5500, area_cost_coeff_exchanger=150,
            area_cost_coeff_heater=150, area_cost_coeff_cooler=150,
            cost_exponent_exchanger=0,
            hot_util_film_coeff=1, cold_util_film_coeff=1,
            decimals=2, disp=True):
    """
    Perform heat exchanger network (HEN) synthesis by Yee and Grossmann's
        transshipment method.

    Notes:
        This model designs a heat exchanger network operating at minimal
            annual cost and satisfying specified heating and cooling
            requirements. The superstructure consists of k stages and up
            to 10 possible exchangers.

        The original model was written based on this article:
            Yee, T F, and Grossmann, I E, Simultaneous Optimization of
            Models for Heat Integration: Heat Exchanger Network Synthesis.
            Computers and Chemical Engineering 14, 10 (1990), 1151-1184.

        The GAMS version of this model is available in the GAMS model
            library:
            https://www.gams.com/47/gamslib_ml/libhtml/gamslib_synheat.html

    Example:
        synheat([H1, H2, C1, C2], DT_min=20, decimals=2)

    Args:
        streams: A list of previously created Stream objects.
        DT_min: Minimum approach temperature. Defaults to 10 °C.
        n_stages: Number of stages ``k`` in the superstructure.
        stages: Same as ``n_stages``.
        T_hot_util_in: Hot utility inlet temperature. Defaults to 450 °C.
        T_hot_util_out: Hot utility outlet temperature. Defaults to 450 °C.
        T_cold_util_in: Cold utility inlet temperature Defaults to 20 °C.
        T_cold_util_out Cold utility outlet temperature. Defaults to 20 °C.
        hot_util_cost: Relative cost of hot utility. Defaults to 100 units.
        cold_util_cost: Relative cost of cold utility. Defaults to 100 units.
        exchanger_fixed_charge: Relative fixed price of exchanger. Defaults
            to 5500 units, but could be different in practice.
        area_cost_coeff_exchanger: Area cost coefficient for exchangers.
            Defaults to 150 units.
        area_cost_coeff_heater: Area cost coefficient for heaters.
            Defaults to 150 units.
        area_cost_coeff_cooler: Area cost coefficient for coolers.
            Defaults to 150 units.
        cost_exponent_exchanger: Cost exponent for exchangers.
            Defaults to 0, but could be different in practice.
        hot_util_film_coeff: Individual film heat transfer coefficient
            for hot utility. Defaults to 1.0 J/(s m^2 °C)
        cold_util_film_coeff=1: Individual film heat transfer coefficient
            for cold utility. Defaults to 1.0 J/(s m^2 °C).
        decimals: Number of decimals in output. Defaults to 2.

    Returns:
        A HENDesign instance containing number of stages, streams, hot and
            cold stream temperatures, and heat exchanged between streams.
    """

    if n_stages is not None and stages is not None:
        print(
            f"Both n_stages (= {n_stages}) and stages (= {stages}) " +
            "are specified; using n_stages"
        )
    elif n_stages is None:
        n_stages = stages
    elif n_stages is None and stages is None:
        print('At least one of n_stages or stages must be specified')
    else:
        pass

    hot_streams = [stream for stream in streams if stream.kind == 'hot']
    cold_streams = [stream for stream in streams if stream.kind == 'cold']
    hot_range = [n for n in range(1, len(hot_streams) + 1)]
    cold_range = [n for n in range(1, len(cold_streams) + 1)]

    m = gams.Container()

    # Sets for streams
    i = gams.Set(m, name='i', records=hot_range, description='hot streams')
    j = gams.Set(m, name='j', records=cold_range, description='cold streams')

    # Sets for stages
    k = gams.Set(
        m, name='k', records=range(1, n_stages + 1),
        description='Temperature locations')
    st = gams.Set(m, 'st', domain=k)  # up to second-last
    firstK = gams.Set(
        m, name='firstK', domain=k, is_singleton=True,
        description='First temperature location'
    )
    lastK = gams.Set(
        m, name='lastK', domain=k, is_singleton=True,
        description='Last temperature location',
    )
    st[k].where[~k.last] = True
    firstK[k].where[k.first] = True
    lastK[k].where[k.last] = True

    # Parameters
    fh = gams.Parameter(
        m, name='fh', domain=i,
        description='Hot stream heat capacity'
    )
    fc = gams.Parameter(
        m, name='fc', domain=j,
        description='Cold stream heat capacity'
    )
    thin = gams.Parameter(
        m, name='thin', domain=i,
        description='Hot stream source temperature')
    thout = gams.Parameter(
        m, name='thout', domain=i,
        description='Hot stream target temperature'
    )
    tcin = gams.Parameter(
        m, name='tcin', domain=j,
        description='Cold stream source temperature'
    )
    tcout = gams.Parameter(
        m, name='tcout', domain=j,
        description='Cold stream target temperature'
    )
    ech = gams.Parameter(
        m, name='ech', domain=i,
        description='Enthalpy content of hot stream i'
    )
    ecc = gams.Parameter(
        m, name='ecc', domain=j,
        description='Enthalpy content of cold stream j'
    )
    hh = gams.Parameter(
        m, name='hh', domain=i,
        description='Individual film coefficient of hot stream i'
    )
    hc = gams.Parameter(
        m, name='hc', domain=j,
        description='Individual film coefficient of cold stream j'
    )
    hucost = gams.Parameter(
        m, name='hucost', domain=[],
        description='Cost of hot utility'
    )
    cucost = gams.Parameter(
        m, name='cucost', domain=[],
        description='Cost of cold utility'
    )
    unitc = gams.Parameter(
        m, name='unitc', domain=[],
        description='Fixed charge for exchanger'
    )
    acoeff = gams.Parameter(
        m, name='acoeff', domain=[],
        description='Area cost coefficient for exchangers'
    )
    hucoeff = gams.Parameter(
        m, name='hucoeff', domain=[],
        description='Area cost coefficient for heaters'
    )
    cucoeff = gams.Parameter(
        m, name='cucoeff', domain=[],
        description='Area cost coefficient for coolers'
    )
    aexp = gams.Parameter(
        m, name='aexp', domain=[],
        description='Cost exponent for exchangers'
    )
    hhu = gams.Parameter(
        m, name='hhu', domain=[],
        description='Individual film coefficient of hot utility'
    )
    hcu = gams.Parameter(
        m, name='ccu', domain=[],
        description='Individual film coefficient of cold utility'
    )
    thuin = gams.Parameter(
        m, name='thuin', domain=[],
        description='Inlet temperature of hot utility'
    )
    thuout = gams.Parameter(
        m, name='thuout', domain=[],
        description='Outlet temperature of cold utility'
    )
    tcuin = gams.Parameter(
        m, name='tcuin', domain=[],
        description='Inlet temperature of hot utility'
    )
    tcuout = gams.Parameter(
        m, name='tcuout', domain=[],
        description='Outlet temperature of cold utility'
    )
    gamma = gams.Parameter(
        m, name='gamma', domain=[i, j],
        description='Upper bound of driving force'
    )
    a = gams.Parameter(
        m, name='a', domain=[i, j, k],
        description='Area for exchanger for match ij in interval k'
    )
    al = gams.Parameter(
        m, name='al', domain=[i, j, k],
        description='Area calculated with log mean'
    )
    acu = gams.Parameter(
        m, name='acu', domain=i,
        description='Area coolers'
    )
    ahu = gams.Parameter(
        m, name='ahu', domain=j,
        description='Area heaters')
    tmapp = gams.Parameter(
        m, name='tmapp', domain=[],
        description='Minimum approach temperature'
    )
    costheat = gams.Parameter(
        m, name='costheat', domain=[],
        description='Cost of heating'
    )
    costcool = gams.Parameter(
        m, name='costcool', domain=[],
        description='Cost of cooling'
    )
    invcost = gams.Parameter(
        m, name='invcost', domain=[],
        description='Investment cost'
    )

    # Variables
    z = gams.Variable(m, name='z', domain=[i, j, k], type='binary')
    zcu = gams.Variable(m, name='zcu', domain=[i], type='binary')
    zhu = gams.Variable(m, name='zhu', domain=[j], type='binary')
    th = gams.Variable(
        m, name='th', domain=[i, k], type='positive',
        description='Temperature of hot stream i as it enters stage k'
    )
    tc = gams.Variable(
        m, name='tc', domain=[j, k], type='positive',
        description='Temperature of cold stream j as it leaves stage k'
    )
    q = gams.Variable(
        m, name='q', domain=[i, j, k], type='positive',
        description=(
            'Energy exchanged between hot stream i ' +
            'and cold stream j in stage k'
        )
    )
    qc = gams.Variable(
        m, name='qc', domain=[i], type='positive',
        description=(
            'Energy exchanged between hot stream i and the cold utility'
        )
    )
    qh = gams.Variable(
        m, name='qh', domain=[j], type='positive',
        description=(
            'Energy exchanged between cold stream j and the hot utility'
        )
    )
    dt = gams.Variable(
        m, name='dt', domain=[i, j, k], type='positive',
        description=(
            'Approach between hot stream i and cold stream j at location k'
        )
    )
    dtcu = gams.Variable(
        m, name='dtcu', domain=[i], type='positive',
        description='Approach between hot stream i and the cold utility'
    )
    dthu = gams.Variable(
        m, name='dthu', domain=[j], type='positive',
        description='Approach between hot stream j and the hot utility'
    )
    cost = gams.Variable(
        m, name='cost',
        description='HEN and utility cost'
    )

    # Equations
    eh = gams.Equation(
        m, name='eh', domain=[i, k],
        description='Energy exchanged by hot stream i in stage k'
    )
    eqc = gams.Equation(
        m, name='eqc', domain=[i],
        description='Energy exchanged by hot stream i with the cold utility'
    )
    teh = gams.Equation(
        m, name='teh', domain=[i],
        description='Total energy exchanged by hot stream i'
    )
    ec = gams.Equation(
        m, name='ec', domain=[j, k],
        description='Energy exchanged by cold stream j in stage k'
    )
    eqh = gams.Equation(
        m, name='eqh', domain=[j],
        description='Energy exchanged by cold stream j with the hot utility'
    )
    tec = gams.Equation(
        m, name='tec', domain=[j],
        description='Total energy exchanged by cold stream j'
    )
    month = gams.Equation(
        m, name='month', domain=[i, k],
        description='Monotonicity of th'
    )
    montc = gams.Equation(
        m, name='montc', domain=[j, k],
        description='Monotonicity of tc'
    )
    monthl = gams.Equation(
        m, name='monthl', domain=[i],
        description='Monotonicity of th k = last'
    )
    montcf = gams.Equation(
        m, name='montcf', domain=[j],
        description='Monotonicity of tc for k = 1'
    )
    tinh = gams.Equation(
        m, name='tinh', domain=[i],
        description='Supply temperature of hot streams'
    )
    tinc = gams.Equation(
        m, name='tinc', domain=[j],
        description='Supply temperature of cold streams'
    )
    logq = gams.Equation(
        m, name='logq', domain=[i, j, k],
        description='Logical constraints on q'
    )
    logqh = gams.Equation(
        m, name='logqh', domain=[j],
        description='Logical constraints on qh(j)'
    )
    logqc = gams.Equation(
        m, name='logqc', domain=[i],
        description='Logical constraints on qc(i)'
    )
    logdth = gams.Equation(
        m, name='logdth', domain=[i, j, k],
        description='Logical constraints on dt at the hot end'
    )
    logdtc = gams.Equation(
        m, name='logdtc', domain=[i, j, k],
        description='Logical constraints on dt at the cold end'
    )
    logdtcu = gams.Equation(
        m, name='logdtcu', domain=[i],
        description='Logical constraints on dtcu'
    )
    logdthu = gams.Equation(
        m, name='logdthu', domain=[j],
        description='Logical constraints on dthu'
    )
    obj = gams.Equation(
        m, name='obj',
        description='Objective function'
    )

    # Equation listing
    teh[i] = (
        (thin[i] - thout[i]) * fh[i]
        == gams.Sum((j, st), q[i, j, st]) + qc[i]
    )
    tec[j] = (
        (tcout[j] - tcin[j]) * fc[j]
        == gams.Sum((i, st), q[i, j, st]) + qh[j]
    )
    eh[i, k].where[st[k]] = (
        fh[i] * (th[i, k] - th[i, k + 1])
        == gams.Sum(j, q[i, j, k])
    )
    ec[j, k].where[st[k]] = (
        fc[j] * (tc[j, k] - tc[j, k + 1])
        == gams.Sum(i, q[i, j, k])
    )
    eqc[i] = (
        fh[i] * (th[i, lastK] - thout[i]) == qc[i]
    )
    eqh[j] = (
        fc[j] * (tcout[j] - tc[j, firstK]) == qh[j]
    )
    tinh[i] = (
        thin[i] == th[i, firstK]
    )
    tinc[j] = (
        tcin[j] == tc[j, lastK]
    )
    month[i, k].where[st[k]] = (
        th[i, k] >= th[i, k + 1]
    )
    montc[j, k].where[st[k]] = (
        tc[j, k] >= tc[j, k + 1]
    )
    monthl[i] = (
        th[i, lastK] >= thout[i]
    )
    montcf[j] = (
        tcout[j] >= tc[j, firstK]
    )
    logq[i, j, k].where[st[k]] = (
        q[i, j, k] - gams.math.Min(ech[i], ecc[j]) * z[i, j, k] <= 0
    )
    logqc[i] = (
        qc[i] - ech[i] * zcu[i] <= 0
    )
    logqh[j] = (
        qh[j] - ecc[j] * zhu[j] <= 0
    )
    logdth[i, j, k].where[st[k]] = (
        dt[i, j, k]
        <= th[i, k] - tc[j, k] + gamma[i, j] * (1 - z[i, j, k])
    )
    logdtc[i, j, k].where[st[k]] = (
        dt[i, j, k + 1]
        <= th[i, k + 1] - tc[j, k + 1] + gamma[i, j] * (1 - z[i, j, k])
    )
    logdthu[j] = (
        dthu[j] <= (thuout - tc[j, firstK])
    )
    logdtcu[i] = (
        dtcu[i] <= th[i, lastK] - tcuout
    )
    obj[...] = cost == (
        # unit capital cost term
        unitc * (
            gams.Sum((i, j, st), z[i, j, st])
            + gams.Sum(i, zcu[i])
            + gams.Sum(j, zhu[j])
        )
        # exchanger area cost (Chen-type) for process stream exchangers
        + acoeff * gams.Sum(
            (i, j, k),
            (
                q[i, j, k] * ((1 / hh[i]) + (1 / hc[j]))
                / (
                    (
                        dt[i, j, k]
                        * dt[i, j, k + 1]
                        * (dt[i, j, k] + dt[i, j, k + 1]) / 2
                        + 1e-6
                    ) ** 0.33333
                    + 1e-6
                )
                + 1e-6
            ) ** aexp,
        )
        # area cost for heaters
        + hucoeff * gams.Sum(
            j,
            (
                qh[j] * ((1 / hc[j]) + 1 / hhu)
                / (
                    (
                        (thuin - tcout[j])
                        * dthu[j]
                        * ((thuin - tcout[j] + dthu[j]) / 2)
                        + 1e-6
                    ) ** 0.33333
                )
                + 1e-6
            ) ** aexp,
        )
        # area cost for coolers
        + cucoeff * gams.Sum(
            i,
            (
                qc[i] * ((1 / hh[i]) + 1 / hcu)
                / (
                    (
                        (thout[i] - tcuin)
                        * dtcu[i]
                        * ((thout[i] - tcuin + dtcu[i]) / 2)
                        + 1e-6
                    ) ** 0.33333
                )
                + 1e-6
            ) ** aexp,
        )
        # utility operating costs
        + gams.Sum(j, qh[j] * hucost)
        + gams.Sum(i, qc[i] * cucost)
    )

    # Process stream data
    ii = 0
    for stream in hot_streams:
        thin[ii + 1] = hot_streams[ii].T_source
        thout[ii + 1] = hot_streams[ii].T_target
        fh[ii + 1] = hot_streams[ii].C
        hh[ii + 1] = hot_streams[ii].h
        ii += 1

    jj = 0
    for stream in cold_streams:
        tcin[jj + 1] = cold_streams[jj].T_source
        tcout[jj + 1] = cold_streams[jj].T_target
        fc[jj + 1] = cold_streams[jj].C
        hc[jj + 1] = cold_streams[jj].h
        jj += 1

    thuin[...] = T_hot_util_in
    thuout[...] = T_hot_util_out
    tcuin[...] = T_cold_util_in
    tcuout[...] = T_cold_util_out
    hucost[...] = hot_util_cost
    cucost[...] = cold_util_cost
    unitc[...] = exchanger_fixed_charge
    acoeff[...] = area_cost_coeff_exchanger
    hucoeff[...] = area_cost_coeff_heater
    cucoeff[...] = area_cost_coeff_cooler
    aexp[...] = cost_exponent_exchanger
    hhu[...] = hot_util_film_coeff
    hcu[...] = cold_util_film_coeff
    tmapp[...] = DT_min

    # Bounds
    dt.lo[i, j, k] = tmapp
    dthu.lo[j] = tmapp
    dtcu.lo[i] = tmapp

    th.up[i, k] = thin[i]
    th.lo[i, k] = thout[i]

    tc.up[j, k] = tcout[j]
    tc.lo[j, k] = tcin[j]

    # Initialization
    th.l[i, k] = thin[i]
    tc.l[j, k] = tcin[j]

    dthu.l[j] = thuout - tcin[j]
    dtcu.l[i] = thin[i] - tcuout

    ech[i] = fh[i] * (thin[i] - thout[i])
    ecc[j] = fc[j] * (tcout[j] - tcin[j])

    gamma[i, j] = gams.math.Max(
        0,
        tcin[j] - thin[i],
        tcin[j] - thout[i],
        tcout[j] - thin[i],
        tcout[j] - thout[i],
    )

    dt.l[i, j, k] = thin[i] - tcin[j]

    # Note: using subset st in the domain instead of a $ condition
    q.l[i, j, st] = gams.math.Min(ech[i], ecc[j])

    # Model and solution
    super_model = gams.Model(
        m,
        name='super',
        equations=m.getEquations(),
        problem=gams.Problem.MINLP,
        sense=gams.Sense.MIN,
        objective=cost,
    )

    solve_options = gams.Options.fromGams(
        {
            'optcr': 0,
            'limrow': 0,
            'limcol': 0,
            'solprint': 0,
            'sysout': 'off',
        }
    )

    super_model.solve(options=solve_options)

    # Post-processing: areas by Chen approximation
    a[i, j, k].where[st[k]] = (
        q.l[i, j, k] * ((1 / hh[i]) + (1 / hc[j]))
        / (
            2 / 3 * (dt.l[i, j, k] * dt.l[i, j, k + 1]) ** 0.5
            + 1 / 6 * (1e-8 + dt.l[i, j, k] + dt.l[i, j, k + 1])
        )
    )

    al[i, j, k].where[st[k]] = (
        q.l[i, j, k] * ((1 / hh[i]) + (1 / hc[j]))
        / (
            dt.l[i, j, k]
            * dt.l[i, j, k + 1]
            * (dt.l[i, j, k] + dt.l[i, j, k + 1]) / 2
        ) ** 0.33333
    )

    # Areas of heaters and coolers
    ahu[j] = (
        qh.l[j]
        * ((1 / hc[j]) + 1 / hhu)
        / (
            (
                (thuin - tcout[j])
                * dthu.l[j]
                * ((thuin - tcout[j] + dthu.l[j]) / 2)
            )
            + 1e-6
        ) ** 0.33333
    )

    acu[i] = (
        qc.l[i]
        * ((1 / hh[i]) + 1 / hcu)
        / (
            (
                (thout[i] - tcuin)
                * dtcu.l[i]
                * ((thout[i] - tcuin + dtcu.l[i]) / 2)
            )
            + 1e-6
        ) ** 0.33333
    )

    # Utility costs
    costheat[...] = gams.Sum(j, qh.l[j] * hucost)
    costcool[...] = gams.Sum(i, qc.l[i] * cucost)

    # Investment cost
    invcost[...] = cost.l - costheat - costcool

    if q is not None:
        Q_exchange=q.l.records.sort_values(by=["k"])
    else:
        Q_exchange=None

    hen = HENDesign(
        streams=streams,
        n_stages=n_stages,
        Q_exchange=Q_exchange,
        T_hot_streams=th,
        T_cold_streams=tc,
    )

    # Print with GAMS-like DISPLAY behavior:
    if disp is True:
        with pd.option_context('display.precision', decimals):
            for var in (th, tc):
                print(f'{var.description}:')
                var_records_round = var.l.records.round(decimals)
                print(f'{var_records_round.to_string(index=False)}\n')

            print(f'{q.description}:')
            q_sorted_by_k = q.l.records.sort_values(by=["k"])
            q_sorted_by_k_round = q_sorted_by_k.round(decimals)
            print(f'{q_sorted_by_k_round.to_string(index=False)}\n')

            for var in (qc, qh, teh, tec):
                print(f'{var.description}:')
                var_records_round = var.l.records.round(decimals)
                print(f'{var_records_round.to_string(index=False)}\n')

            for var in (costheat, costcool, invcost):
                print(f'{var.description}: {var.toValue():.{decimals}f}')

    return hen


syn_heat = synheat
heat_integrate = synheat
synthesize_hen = hen_synthesize = synheat
design_hen = hen_design = synheat
grossmann = yee_grossmann = yee_and_grossmann = synheat



import pickle


def duplicate_figure(fig):
    """
    Duplicates a figure.

    Example:
        fignew, axnew = duplicate(fig)

    Args:
        fig: A previously created figure object.

    Returns:
        A figure object that is a duplicate of the original figure and a
            corresponding axes object.
    """
    pkl = pickle.dumps(fig)  # serialize figure
    fignew = pickle.loads(pkl)  # deserialize into fignew
    axnew = fignew.axes[0]
    return fignew, axnew



from .simplex import simplex_tableaux
from .desmos import desmos_graph
from .ReactionSystem import *
from .Reactors import *
from .figures import *
from .TernarySystem import *
from .Stream import *
from .synheat import *


# Reaction Rate Field Plot: Reaction Systems
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025, originally written in MATLAB in 2017
# Licensed under GNU Public License 3.0
# Author contact info: gsriram@umd.edu

import numpy as np
import matplotlib.pyplot as plt


class ReactionSystem:
    """
    Container for components and reactions in a reactive system.

    Example:
        system = ReactionSystem(
            component_ids=('A', 'B', 'C'),
            component_names={'A': 'n-pentane',
                             'B': '2-methylbutane',
                             'C': '2,2-dimethylpropane'},
            axes={'A': 0, 'B': 1},
            h_lim=[0, 1],
            v_lim=[0, 1],
            reactions=('r1', 'r2'),
            stoich={'r1': {'A': -1, 'B': 1},
                    'r2': {'B': -1, 'C': 1}},
            kinetics={'r1': lambda C: 3 * C['A'],
                      'r2': lambda C: 2 * C['B']},
            inequality=lambda C: C['A'] + C['B'] - 1
        )

    Args:
        component_ids: Short ids (abbreviations or formulas) for components.
            Enclose component_ids in parentheses.
        component_names: Names or full names of components.
        axes: Dictionary indicating which component is represented on which
            axis. In the example above, component 'A' is represented on axis
            0 (x-axis) and component B is represented on axis 1 (y-axis).
        h_lim (or x_lim): Horizontal axis limits.
        v_lim (or y_lim): Vertical axis limits.
        reactions: Reaction ids. Enclose in parentheses.
        stoich: Reaction stoichiometry. Reactants get negative coefficients,
            and products get positive coefficients.
        kinetics: Reaction kinetics. Kinetics should be expressed in terms of
            the state variables such as concentrations, mole fractions,
            number of moles or partial pressures. Kinetics could be expressed
            by lambda functions as shown in the example above, or in long form
            as follows:

                k1 = 3  # s^-1
                k2 = 2  # s^-1
                def r1(C):
                    return k1 * C['A']
                def r2(C):
                    return k2 * C['B']
                kinetics = {'r1': r1,
                            'r2': r2}

        inequality: Inequality constraints that constrain the state variables.
            These constraints usually result from material balances, mole
            fraction or partial pressure summations, etc. The inequality
            constraints should be expressed as the left side of a vector that
            should be <= 0, for example:

                inequality=lambda C: [C['A'] + C['B'] - 1, C['A'] - 0.5]

            If the vector has only one element, as in the example above,
            express it as a scalar without the brackets:

                inequality = lambda C: C['A'] + C['B'] - 1

            If there are no inequality constraints, express as:

                inequality=lambda C: None

    Returns:
        A ReactionSystem object containing the specified components, reactions,
            stoichiometry, kinetics, and inequality constraints.
    """

    def __init__(self, component_ids, component_names, axes,
                 reactions, kinetics, stoich=None, stoichiometry=None,
                 inequality=lambda x: None,
                 h_lim=None, x_lim=None, v_lim=None, y_lim=None):
        self.component_ids = component_ids
        self.components = self.component_ids
        self.component_names = component_names
        self.names = self.component_names
        self.axes = axes

        if h_lim is None and x_lim is not None:
            h_lim = x_lim
        elif h_lim is not None and x_lim is not None:
            raise ValueError("Use either h_lim or x_lim, not both")
        if h_lim is None:
            h_lim = [0, 1]
        self.h_lim = h_lim

        if v_lim is None and y_lim is not None:
            v_lim = y_lim
        elif v_lim is not None and y_lim is not None:
            raise ValueError('Use either v_lim or y_lim, not both')
        if v_lim is None:
            v_lim = [0, 1]
        self.v_lim = v_lim

        if 0 not in list(self.axes.values()):
            raise KeyError('Axis #0 (x-axis) must be specified')
        if 1 not in list(self.axes.values()):
            raise KeyError('Axis #1 (y-axis) must be specified')

        self.reactions = reactions

        if stoich is not None and stoichiometry is None:
            self.stoich = stoich
        if stoich is None and stoichiometry is not None:
            self.stoich = stoichiometry
        if stoich is not None and stoichiometry is not None:
            raise ValueError('Specify stoich or stoichiometry, not both')
        if stoich is None and stoichiometry is None:
            raise ValueError('Stoichiometry must be specified')
        if not set(self.reactions).issubset(self.stoich):
            raise KeyError('Stoichiometry must be specified for all reactions')

        self.kinetics = kinetics

        if not set(self.reactions).issubset(self.kinetics):
            raise KeyError('Kinetics must be specified for all reactions')

        self.inequality = inequality

    def __repr__(self):
        rows_stoich = '\n'.join(f'        {rxn}: {s}'
                                for rxn, s in self.stoich.items())
        rows_kinetics = '\n'.join(f'        {rxn}: {k}'
                                  for rxn, k in self.kinetics.items())
        disp_text = (f'Reaction system:\n'
                     f'    component ids:\n\t{self.components}\n'
                     f'    component names:\n\t{self.component_names}\n'
                     f'    axes:\n\t{self.axes}\n'
                     f'    horizontal axis limits:\n\t{self.h_lim}\n'
                     f'    vertical axis limits:\n\t{self.v_lim}\n'
                     f'    reactions:\n\t{self.reactions}\n'
                     f'    stoichiometry:\n{rows_stoich}\n'
                     f'    kinetics:\n{rows_kinetics}\n'
                     f'    inequalities:\n\t{self.inequality}\n')
        return disp_text

    def rate(self, x):
        """
        Rates of formation of state variables at a particular value of the
            state vector x.

        Example:
            system.rate([0.4, 0.3])

        Args:
            x: State vector, which could represent concentration, mole
                fraction, number of moles or partial pressure.

        Returns:
            Array consisting of the rates of formation of the state variables.
        """

        x = np.asarray(x)
        if not np.issubdtype(x.dtype, np.number):
            raise TypeError(
                'x must be an array with integer or float elements'
                )
        if min(x.shape) < 2:
            x_size_error = 'x must be an array with two components'
            raise ValueError(x_size_error)

        X = {'A': 0, 'B': 0}  # X is the concentration vector (or equivalent)
        r_rxn = dict.fromkeys(self.reactions, 0)
        r_comp = dict.fromkeys(self.axes, 0)

        for i in self.components:
            if i in self.axes:
                X[i] = x[self.axes[i]]

        for j in self.reactions:
            r_rxn[j] += self.kinetics[j](X)

        for i in self.components:
            if i in self.axes:
                for j in self.reactions:
                    if i in self.stoich[j]:
                        r_comp[i] += self.stoich[j][i] * r_rxn[j]

        r_comp = np.array(list(r_comp.values()))
        return r_comp

    def inequality_status(self, x):
        """
        Status of inequality constraints at a particular value of the
            state vector x.

        Example:
            system.inequality_status([0.4, 0.3])

        Args:
            x: State vector, which could represent concentration, mole
                fraction, number of moles or partial pressure.

        Returns:
            Array consisting of the statuses of each inequality constraint.
                A value <= 0 indicates the state vector is feasible, whereas
                a value > 0 indicates it is infeasible as per the specified
                constraints.
        """

        x = np.asarray(x)
        if not np.issubdtype(x.dtype, np.number):
            raise TypeError(
                'x must be an array with integer or float elements'
                )
        if min(x.shape) < 2:
            x_size_error = 'x must be an array with two components'
            raise ValueError(x_size_error)

        X = {'A': 0, 'B': 0}  # X is the concentration vector (or equivalent)

        for i in self.components:
            if i in self.axes:
                X[i] = x[self.axes[i]]

        if self.inequality(X) is None:
            inequality_status = np.array(0)
        else:
            inequality_status = np.array(self.inequality(X))
        return inequality_status

    def plot_rate_field(self,
                        ax=None,
                        n_vec=51, n_vec_h=None, n_vec_v=None,
                        inequality_tol=1e-6,
                        arrow_scale=None, arrow_scale_h=None,
                        arrow_scale_v=None,
                        fsize=12, facecolor='white', font=12,
                        **kwargs):

        """
        Plot rate field in concentration space.

        Example:
            system.draw_rate_field(
                n_vec=51, n_vec_h=51, n_vec_v=51, inequality_tol=1e-6,
                arrow_scale=51, arrow_scale_h=51, arrow_scale_v=51,
                fsize=12, facecolor='white', font=12, **kwargs
            )

        Args:
            n_vec: Number of rate vectors along any dimension, defaults to 51.
            n_vec_h: Number of rate vectors along horizontal dimension
                (x-axis). Defaults to 51, can also be set by n_vec.
            n_vec_v: Number of rate vectors along vertical dimension
                (x-axis). Defaults to 51, can also be set by n_vec.
            inequality_tol: Tolerance for any inequalities, defaults to 1e-6.
            arrow_scale: scale of arrow along any dimension, defaults to n_vec.
            arrow_scale_h: Scale of arrow along horizontal dimension (x-axis).
                Defaults to n_vec_x.
            arrow_scale_v: Scale of arrow along vertical dimension (y-axis).
                Defaults to n_vec_y.
            fsize: Figure size, defaults to 12.
            facecolor: Figure background color, defaults to ``white``.
            font: Base font used for axis labels, etc. Defaults to 12.
            **kwargs: Additional keyword arguments passed to
                matplotlib.pyplot.quiver. Common options include ``color``
                and ``linewidth``.

        Returns:
            Figure: Figure object containing the rate field plot.
            Axes: Associated axes object.
        """

        if ax is None:
            fig, ax = plt.subplots(figsize=(fsize, fsize), facecolor=facecolor)
        else:
            fig = ax.figure
        h_lim = self.h_lim
        v_lim = self.v_lim
        for spine in ax.spines.values():
            spine.set_visible(True)
            spine.set_linewidth(1)
        ax.set_xlim(h_lim)
        ax.set_ylim(v_lim)
        ax.grid(True)

        for i in self.components:
            if i in self.axes:
                if self.axes[i] == 0:
                    ax.set_xlabel(self.names[i], fontsize=font)
                if self.axes[i] == 1:
                    ax.set_ylabel(self.names[i], fontsize=font)

        if n_vec_h is None:
            n_vec_h = n_vec
            arrow_scale_h = n_vec_h
        if n_vec_v is None:
            n_vec_v = n_vec
            arrow_scale_v = n_vec_v

        if arrow_scale is None:
            arrow_scale = n_vec
        else:
            if arrow_scale_h is None:
                arrow_scale_h = arrow_scale
            if arrow_scale_v is None:
                arrow_scale_v = arrow_scale

        # Compute rate vectors
        X = np.linspace(h_lim[0], h_lim[1], n_vec_h)
        Y = np.linspace(v_lim[0], v_lim[1], n_vec_v)

        xx, yy = np.meshgrid(X, Y)
        inequalities = self.inequality_status([xx, yy])
        n_inequalities = np.size(self.inequality_status([0, 0]))

        XY = np.array([xx.ravel(), yy.ravel()])
        rates = self.rate(XY)
        uu = rates[0].reshape(xx.shape)
        vv = rates[1].reshape(yy.shape)

        if n_inequalities == 1:
            if np.any(inequalities != 0):
                uu = np.where(inequalities > inequality_tol, np.nan, uu)
                vv = np.where(inequalities > inequality_tol, np.nan, vv)
        else:
            mask = np.any(inequalities > inequality_tol, axis=0)
            uu = np.where(mask, np.nan, uu)
            vv = np.where(mask, np.nan, vv)

        # Normalize
        mag = np.hypot(uu, vv)
        un, vn = uu, vv
        with np.errstate(invalid='ignore'):
            un = np.where(np.isnan(un), np.nan,
                          np.where(mag > 0,
                                   np.divide(uu, mag), np.nan))
            vn = np.where(np.isnan(vn), np.nan,
                          np.where(mag > 0,
                                   np.divide(vv, mag), np.nan))

        quiver = ax.quiver(X, Y, un, vn, color='#888888', angles='xy',
                           scale=arrow_scale, alpha=0.5, **kwargs)
        # angles='xy' plots from x, y to x+u, y+v

        return fig, ax, quiver

    draw_rate_field = plot_rate_field
    simulate_rate_field = plot_rate_field


ReactiveSystem = ReactionSystem
